@@ -41,6 +41,7 @@
 #include "ui/contact_info.h"
 #include "ui/contacts.h"
 #include "ui/delete_messages.h"
+#include "ui/files.h"
 #include "ui/invite_contact.h"
 #include "ui/messenger.h"
 #include "ui/new_account.h"
@@ -48,6 +49,7 @@
 #include "ui/new_group.h"
 #include "ui/new_lobby.h"
 #include "ui/new_platform.h"
+#include "ui/new_tag.h"
 #include "ui/play_media.h"
 #include "ui/send_file.h"
 #include "ui/settings.h"
@@ -124,19 +126,20 @@ typedef struct MESSENGER_Application
     UI_SEND_FILE_Handle send_file;
     UI_PLAY_MEDIA_Handle play_media;
 
+    UI_NEW_ACCOUNT_Handle new_account;
     UI_NEW_CONTACT_Handle new_contact;
     UI_NEW_GROUP_Handle new_group;
     UI_NEW_LOBBY_Handle new_lobby;
     UI_NEW_PLATFORM_Handle new_platform;
-    UI_NEW_ACCOUNT_Handle new_account;
+    UI_NEW_TAG_Handle new_tag;
 
     UI_ACCOUNTS_Handle accounts;
+    UI_FILES_Handle files;
     UI_CONTACTS_Handle contacts;
     UI_SETTINGS_Handle settings;
   } ui;
 
   struct {
-    gboolean mobile_design;
     gboolean hide_delete_dialog;
 
     gboolean autostart;
@@ -194,6 +197,31 @@ application_get_resource_path(MESSENGER_Application *app,
 void
 application_run(MESSENGER_Application *app);
 
+/**
+ * Initialize the pipewire core of the messenger 
+ * application if possible.
+ *
+ * @param app Messenger application
+ */
+void
+application_pw_core_init(MESSENGER_Application *app);
+
+/**
+ * Cleanups the pipewire core of the messenger
+ * application if available.
+ */
+void
+application_pw_core_cleanup(MESSENGER_Application *app);
+
+/**
+ * Run the pipewire main loop of the messenger 
+ * application if available.
+ *
+ * @param app Messenger application
+ */
+void
+application_pw_main_loop_run(MESSENGER_Application *app);
+
 /**
  * Shows the messenger application main window.
  *
@@ -223,6 +251,17 @@ void
 application_call_event(MESSENGER_Application *app,
                        MESSENGER_ApplicationEvent event);
 
+/**
+ * Calls a given event with the messenger application
+ * syncronously.
+ *
+ * @param app Messenger application
+ * @param event Event
+ */
+void
+application_call_sync_event(MESSENGER_Application *app,
+		                        MESSENGER_ApplicationEvent event);
+
 /**
  * Calls a given message event with the messenger
  * application asyncronously but explicitly synchronized
