@@ -114,10 +114,21 @@ _chat_messenger_message(void *cls,
     }
     case GNUNET_CHAT_KIND_LOGOUT:
     {
-      application_call_event(app, event_cleanup_profile);
+      application_call_sync_event(app, event_cleanup_profile);
       break;
     }
-    case GNUNET_CHAT_KIND_UPDATE:
+    case GNUNET_CHAT_KIND_CREATED_ACCOUNT:
+    case GNUNET_CHAT_KIND_UPDATE_ACCOUNT:
+    {
+      application_call_message_event(
+          app,
+          event_select_profile,
+          context,
+          message
+      );
+      break;
+    }
+    case GNUNET_CHAT_KIND_UPDATE_CONTEXT:
     {
       application_call_message_event(
           app,
@@ -142,6 +153,7 @@ _chat_messenger_message(void *cls,
       break;
     }
     case GNUNET_CHAT_KIND_CONTACT:
+    case GNUNET_CHAT_KIND_SHARED_ATTRIBUTES:
     {
       application_call_message_event(
       	  app,
@@ -196,6 +208,11 @@ _chat_messenger_message(void *cls,
       );
       break;
     }
+    case GNUNET_CHAT_KIND_ATTRIBUTES:
+    {
+      application_call_event(app, event_update_attributes);
+      break;
+    }
     default:
       break;
   }
