@@ -24,13 +24,16 @@
 
 #include "event.h"
 
+#include "account.h"
 #include "application.h"
 #include "contact.h"
 #include "file.h"
 #include "ui.h"
 
 #include "ui/chat_entry.h"
+#include "ui/chat_title.h"
 #include "ui/message.h"
+
 #include <gnunet/gnunet_chat_lib.h>
 #include <gnunet/gnunet_common.h>
 
@@ -59,8 +62,10 @@ _show_notification(MESSENGER_Application *app,
                    const char *category)
 {
   g_assert(
-    (app) && (context) && (contact) && 
-    (text) && (icon) && (category)
+    (app) &&
+    (text) &&
+    (icon) &&
+    (category)
   );
 
   if (app->settings.disable_notifications)
@@ -94,7 +99,7 @@ event_handle_warning(MESSENGER_Application *app,
 		                 struct GNUNET_CHAT_Context *context,
 		                 const struct GNUNET_CHAT_Message *msg)
 {
-  g_assert((app) && (context) && (msg));
+  g_assert((app) && (msg));
 
   const char *text = GNUNET_CHAT_message_get_text(msg);
 
@@ -114,6 +119,42 @@ event_handle_warning(MESSENGER_Application *app,
   );
 }
 
+static enum GNUNET_GenericReturnValue
+_iterate_reload_account(void *cls,
+                        UNUSED const struct GNUNET_CHAT_Handle *handle,
+                        struct GNUNET_CHAT_Account *account)
+{
+  g_assert((cls) && (account));
+
+  MESSENGER_Application *app = (MESSENGER_Application*) cls;
+
+  account_create_info(account);
+  account_update_attributes(account, app);
+
+  return GNUNET_YES;
+}
+
+static void
+_reload_accounts(MESSENGER_Application *app)
+{
+  UI_MESSENGER_Handle *ui = &(app->ui.messenger);
+  CHAT_MESSENGER_Handle *chat = &(app->chat.messenger);
+
+  GNUNET_CHAT_iterate_accounts(
+    chat->handle,
+    _iterate_reload_account,
+    app
+  );
+
+  if (gtk_widget_is_visible(GTK_WIDGET(ui->main_window)))
+    ui_messenger_refresh(app, ui);
+  else
+    ui_accounts_dialog_refresh(app, &(app->ui.accounts));
+
+  if (GNUNET_CHAT_get_connected(chat->handle))
+    application_show_window(app);
+}
+
 static gboolean
 _idle_refresh_accounts(gpointer user_data)
 {
@@ -121,16 +162,13 @@ _idle_refresh_accounts(gpointer user_data)
 
   MESSENGER_Application *app = (MESSENGER_Application*) user_data;
 
-  if (!(app->ui.messenger.main_window))
-    goto refresh_exit;
-
-  if (gtk_widget_is_visible(GTK_WIDGET(app->ui.messenger.main_window)))
-    ui_messenger_refresh(app, &(app->ui.messenger));
-  else
-    ui_accounts_dialog_refresh(app, &(app->ui.accounts));
-
-refresh_exit:
   app->ui.messenger.account_refresh = 0;
+
+  if (!(app->ui.messenger.main_window))
+    return FALSE;
+
+  _reload_accounts(app);
+
   return FALSE;
 }
 
@@ -140,10 +178,10 @@ event_refresh_accounts(MESSENGER_Application *app)
   g_assert(app);
 
   if (app->ui.messenger.account_refresh)
-    g_source_remove(app->ui.messenger.account_refresh);
+    util_source_remove(app->ui.messenger.account_refresh);
 
   if (app->ui.messenger.main_window)
-    app->ui.messenger.account_refresh = g_idle_add(
+    app->ui.messenger.account_refresh = util_idle_add(
       G_SOURCE_FUNC(_idle_refresh_accounts),
       app
     );
@@ -195,14 +233,7 @@ _idle_chat_entry_update(gpointer user_data)
       (!(entry->chat->send_text_view)))
     goto update_exit;
 
-  struct GNUNET_CHAT_Context *context = (struct GNUNET_CHAT_Context*) (
-    g_object_get_qdata(
-      G_OBJECT(entry->chat->send_text_view),
-      entry->chat->app->quarks.data
-    )
-  );
-
-  ui_chat_entry_update(entry, entry->chat->app, context);
+  ui_chat_entry_update(entry, entry->chat->app);
 
 update_exit:
   entry->update = 0;
@@ -215,9 +246,9 @@ enqueue_chat_entry_update(UI_CHAT_ENTRY_Handle *entry)
   g_assert(entry);
 
   if (entry->update)
-    g_source_remove(entry->update);
+    util_source_remove(entry->update);
 
-  entry->update = g_idle_add(
+  entry->update = util_idle_add(
     G_SOURCE_FUNC(_idle_chat_entry_update),
     entry
   );
@@ -230,22 +261,20 @@ _add_new_chat_entry(MESSENGER_Application *app,
   g_assert((app) && (context));
 
   UI_MESSENGER_Handle *ui = &(app->ui.messenger);
-  UI_CHAT_ENTRY_Handle *entry = ui_chat_entry_new(app);
+  UI_CHAT_ENTRY_Handle *entry = ui_chat_entry_new(app, context);
 
   enqueue_chat_entry_update(entry);
 
   gtk_container_add(GTK_CONTAINER(ui->chats_listbox), entry->entry_box);
-  GNUNET_CHAT_context_set_user_pointer(context, entry);
 
   gtk_container_add(
     GTK_CONTAINER(ui->chats_stack),
     entry->chat->chat_box
   );
 
-  g_object_set_qdata(
-    G_OBJECT(entry->chat->send_text_view),
-    app->quarks.data,
-    context
+  gtk_container_add(
+    GTK_CONTAINER(ui->chat_title_stack),
+    entry->chat->title->chat_title_box
   );
 
   ui->chat_entries = g_list_append(ui->chat_entries, entry);
@@ -259,9 +288,9 @@ _add_new_chat_entry(MESSENGER_Application *app,
   );
 
   if (ui->chat_selection)
-    g_source_remove(ui->chat_selection);
+    util_source_remove(ui->chat_selection);
 
-  ui->chat_selection = g_idle_add(
+  ui->chat_selection = util_idle_add(
     G_SOURCE_FUNC(_select_chat_to_activate),
     entry
   );
@@ -307,6 +336,45 @@ _iterate_profile_groups(void *cls,
   return GNUNET_YES;
 }
 
+void
+event_update_profile(MESSENGER_Application *app)
+{
+  g_assert(app);
+
+  if (app->ui.new_account.dialog)
+    ui_new_account_dialog_update(app, &(app->ui.new_account));
+
+  UI_MESSENGER_Handle *ui = &(app->ui.messenger);
+  CHAT_MESSENGER_Handle *chat = &(app->chat.messenger);
+
+  _reload_accounts(app);
+
+  const char *name = GNUNET_CHAT_get_name(chat->handle);
+
+  account_switch_name_avatar_to_info(
+    GNUNET_CHAT_get_connected(chat->handle),
+    ui->profile_avatar
+  );
+
+  ui_label_set_text(ui->profile_label, name);
+
+  const char *key = GNUNET_CHAT_get_key(chat->handle);
+
+  ui_label_set_text(ui->profile_key_label, key);
+
+  gtk_stack_set_visible_child(ui->chats_stack, ui->no_chat_box);
+  
+  GList *children = gtk_container_get_children(GTK_CONTAINER(ui->leaflet_chat));
+
+  if (children) {
+    hdy_leaflet_set_visible_child(ui->leaflet_chat, GTK_WIDGET(children->data));
+    g_list_free(children);
+  }
+
+  GNUNET_CHAT_iterate_contacts(chat->handle, _iterate_profile_contacts, app);
+  GNUNET_CHAT_iterate_groups(chat->handle, _iterate_profile_groups, app);
+}
+
 static void
 _clear_chat_entry(GtkWidget *widget,
 		              gpointer user_data)
@@ -328,65 +396,69 @@ _clear_chat_entry(GtkWidget *widget,
 
   ui->chat_entries = g_list_remove(ui->chat_entries, entry);
 
-  g_object_set_qdata(
-    G_OBJECT(entry->chat->send_text_view),
-    app->quarks.data,
-    NULL
-  );
-
   ui_chat_entry_dispose(entry, app);
 }
 
-void
-event_update_profile(MESSENGER_Application *app)
-{
-  g_assert(app);
-
-  UI_MESSENGER_Handle *ui = &(app->ui.messenger);
-  CHAT_MESSENGER_Handle *chat = &(app->chat.messenger);
-
-  const char *name = GNUNET_CHAT_get_name(chat->handle);
-
-  ui_avatar_set_text(ui->profile_avatar, name);
-  ui_label_set_text(ui->profile_label, name);
-
-  const char *key = GNUNET_CHAT_get_key(chat->handle);
-
-  ui_label_set_text(ui->profile_key_label, key);
-
-  GList *entries = gtk_container_get_children(
-    GTK_CONTAINER(ui->chats_listbox)
-  );
-
-  g_list_foreach(entries, (GFunc) _clear_chat_entry, app);
-  g_list_free(entries);
-
-  gtk_stack_set_visible_child(ui->chats_stack, ui->no_chat_box);
-
-  GNUNET_CHAT_iterate_contacts(chat->handle, _iterate_profile_contacts, app);
-  GNUNET_CHAT_iterate_groups(chat->handle, _iterate_profile_groups, app);
-}
-
 static int
 _cleanup_profile_contacts(void *cls,
 			                    UNUSED struct GNUNET_CHAT_Handle *handle,
 			                    struct GNUNET_CHAT_Contact *contact)
 {
-  g_assert((cls) && (contact));
-
   if (contact)
     contact_destroy_info(contact);
   return GNUNET_YES;
 }
 
+static int
+_cleanup_profile_files(void *cls,
+			                    UNUSED struct GNUNET_CHAT_Handle *handle,
+			                    struct GNUNET_CHAT_File *file)
+{
+  if (file)
+    file_destroy_info(file);
+  return GNUNET_YES;
+}
+
 void
 event_cleanup_profile(MESSENGER_Application *app)
 {
   g_assert(app);
 
+  UI_MESSENGER_Handle *ui = &(app->ui.messenger);
   CHAT_MESSENGER_Handle *chat = &(app->chat.messenger);
 
+  account_remove_name_avatar_from_info(
+    GNUNET_CHAT_get_connected(chat->handle),
+    ui->profile_avatar
+  );
+
+  GList *entries = gtk_container_get_children(
+    GTK_CONTAINER(ui->chats_listbox)
+  );
+
+  g_list_foreach(entries, (GFunc) _clear_chat_entry, app);
+  g_list_free(entries);
+
   GNUNET_CHAT_iterate_contacts(chat->handle, _cleanup_profile_contacts, NULL);
+  GNUNET_CHAT_iterate_files(chat->handle, _cleanup_profile_files, NULL);
+}
+
+void
+event_select_profile(MESSENGER_Application *app,
+		                 struct GNUNET_CHAT_Context *context,
+		                 const struct GNUNET_CHAT_Message *msg)
+{
+  g_assert((app) && (!context) && (msg));
+
+  UI_MESSENGER_Handle *ui = &(app->ui.messenger);
+  CHAT_MESSENGER_Handle *chat = &(app->chat.messenger);
+
+  const struct GNUNET_CHAT_Account *account = GNUNET_CHAT_message_get_account(msg);
+
+  if (GNUNET_CHAT_KIND_CREATED_ACCOUNT == GNUNET_CHAT_message_get_kind(msg))
+    GNUNET_CHAT_connect(chat->handle, account);
+
+  ui_messenger_refresh(app, ui);
 }
 
 gboolean
@@ -428,7 +500,7 @@ event_update_chats(MESSENGER_Application *app,
       enqueue_chat_entry_update(handle);
 
     if (app->settings.leave_chats_delay > 0)
-      g_timeout_add_seconds(
+      util_timeout_add_seconds(
         app->settings.leave_chats_delay,
         _delayed_context_drop,
         context
@@ -437,7 +509,15 @@ event_update_chats(MESSENGER_Application *app,
   else if ((handle) && (handle->entry_box))
     _clear_chat_entry(gtk_widget_get_parent(handle->entry_box), app);
 
-  contact_create_info(GNUNET_CHAT_message_get_sender(msg));
+  struct GNUNET_CHAT_Contact *contact = GNUNET_CHAT_message_get_sender(
+    msg
+  );
+
+  if (!contact)
+    return;
+
+  if (GNUNET_YES == contact_create_info(contact))
+    contact_update_attributes(contact, app);
 }
 
 static void
@@ -502,7 +582,9 @@ event_presence_contact(MESSENGER_Application *app,
   message = ui_message_new(app, UI_MESSAGE_STATUS);
   ui_message_update(message, app, msg);
 
+  contact_update_attributes(contact, app);
   _update_contact_context(contact);
+
   ui_message_set_contact(message, contact);
 
   const enum GNUNET_CHAT_MessageKind kind = GNUNET_CHAT_message_get_kind(
@@ -549,7 +631,7 @@ event_update_contacts(UNUSED MESSENGER_Application *app,
 		                  struct GNUNET_CHAT_Context *context,
 		                  const struct GNUNET_CHAT_Message *msg)
 {
-  g_assert((app) && (context) && (msg));
+  g_assert((app) && (msg));
 
   struct GNUNET_CHAT_Contact *contact = GNUNET_CHAT_message_get_sender(
     msg
@@ -558,9 +640,15 @@ event_update_contacts(UNUSED MESSENGER_Application *app,
   if (!contact)
     return;
 
+  if (GNUNET_CHAT_KIND_SHARED_ATTRIBUTES == GNUNET_CHAT_message_get_kind(msg))
+    contact_update_attributes(contact, app);
+
   contact_update_info(contact);
   _update_contact_context(contact);
 
+  if (!context)
+    return;
+
   UI_CHAT_ENTRY_Handle *handle = GNUNET_CHAT_context_get_user_pointer(context);
 
   if (!handle)
@@ -753,7 +841,7 @@ event_receive_message(MESSENGER_Application *app,
       "im.received"
     );
 
-  ui_label_set_text(message->text_label, text);
+  ui_label_set_markup_text(message->text_label, text);
   ui_label_set_text(message->timestamp_label, time);
 
   ui_chat_add_message(handle->chat, app, message);
@@ -763,7 +851,9 @@ skip_message:
 }
 
 static void
-_event_update_tag_message_state(const struct GNUNET_CHAT_Message *msg)
+_event_update_tag_message_state(MESSENGER_Application *app,
+                                struct GNUNET_CHAT_Context *context,
+                                const struct GNUNET_CHAT_Message *msg)
 {
   g_assert((msg) && (GNUNET_CHAT_KIND_TAG == GNUNET_CHAT_message_get_kind(msg)));
 
@@ -773,11 +863,40 @@ _event_update_tag_message_state(const struct GNUNET_CHAT_Message *msg)
   if (!target)
     return;
 
-  const struct GNUNET_CHAT_Contact *contact;
+  struct GNUNET_CHAT_Contact *contact;
   contact = GNUNET_CHAT_message_get_sender(target);
 
   if (contact)
     contact_update_info(contact);
+
+  UI_CHAT_ENTRY_Handle *handle = GNUNET_CHAT_context_get_user_pointer(context);
+
+  if ((!handle) || (!(handle->chat)))
+    return;
+
+  GList *rows = gtk_container_get_children(
+    GTK_CONTAINER(handle->chat->messages_listbox)
+  );
+
+  for (GList *row = rows; row; row = row->next)
+  {
+    UI_MESSAGE_Handle *message = (UI_MESSAGE_Handle*) g_object_get_qdata(
+      G_OBJECT(row->data), app->quarks.ui
+    );
+
+    if ((message) && (message->msg == target))
+    {
+      if (GNUNET_YES == GNUNET_CHAT_message_is_deleted(msg))
+        ui_message_remove_tag(message, app, msg);
+      else
+        ui_message_add_tag(message, app, msg);
+
+      break;
+    }
+  }
+
+  if (rows)
+    g_list_free(rows);
 }
 
 void
@@ -813,7 +932,7 @@ event_delete_message(MESSENGER_Application *app,
     g_list_free(rows);
 
   if (GNUNET_CHAT_KIND_TAG == GNUNET_CHAT_message_get_kind(msg))
-    _event_update_tag_message_state(msg);
+    _event_update_tag_message_state(app, context, msg);
 
   enqueue_chat_entry_update(handle);
 }
@@ -829,7 +948,7 @@ event_tag_message(MESSENGER_Application *app,
 
   const struct GNUNET_CHAT_Message *target = GNUNET_CHAT_message_get_target(msg);
 
-  _event_update_tag_message_state(msg);
+  _event_update_tag_message_state(app, context, msg);
 
   if ((!handle) || (!(handle->chat)))
     return;
@@ -856,3 +975,40 @@ event_tag_message(MESSENGER_Application *app,
 
   enqueue_chat_entry_update(handle);
 }
+
+static enum GNUNET_GenericReturnValue
+_iterate_contacts_update_own(void *cls,
+                             UNUSED struct GNUNET_CHAT_Handle *handle,
+                             struct GNUNET_CHAT_Contact *contact)
+{
+  g_assert((cls) && (contact));
+
+  MESSENGER_Application *app = (MESSENGER_Application*) cls;
+
+  if (GNUNET_YES != GNUNET_CHAT_contact_is_owned(contact))
+    return GNUNET_YES;
+
+  contact_update_attributes(contact, app);
+  return GNUNET_YES;
+}
+
+void
+event_update_attributes(MESSENGER_Application *app)
+{
+  g_assert(app);
+
+  CHAT_MESSENGER_Handle *chat = &(app->chat.messenger);
+
+  const struct GNUNET_CHAT_Account *account = GNUNET_CHAT_get_connected(
+    chat->handle
+  );
+
+  if (account)
+    account_update_attributes(account, app);
+
+  GNUNET_CHAT_iterate_contacts(
+    chat->handle,
+    _iterate_contacts_update_own,
+    app
+  );
+}
