@@ -26,16 +26,19 @@
 
 #include <gdk/gdkkeysyms.h>
 #include <gnunet/gnunet_chat_lib.h>
+#include <gnunet/gnunet_common.h>
 #include <gnunet/gnunet_time_lib.h>
 #include <stdlib.h>
 
 #include "chat_entry.h"
+#include "chat_title.h"
+#include "file_entry.h"
 #include "file_load_entry.h"
+#include "media_preview.h"
 #include "message.h"
 #include "messenger.h"
 #include "picker.h"
 #include "account_entry.h"
-#include "delete_messages.h"
 
 #include "../application.h"
 #include "../file.h"
@@ -68,26 +71,12 @@ handle_chat_details_via_button_click(UNUSED GtkButton* button,
   UI_CHAT_Handle *handle = (UI_CHAT_Handle*) user_data;
 
   gtk_widget_set_sensitive(GTK_WIDGET(handle->messages_listbox), FALSE);
-  g_idle_add(
+  util_idle_add(
     G_SOURCE_FUNC(_flap_chat_details_reveal_switch),
     handle
   );
 }
 
-static void
-handle_popover_via_button_click(UNUSED GtkButton *button,
-				                        gpointer user_data)
-{
-  g_assert(user_data);
-
-  GtkPopover *popover = GTK_POPOVER(user_data);
-
-  if (gtk_widget_is_visible(GTK_WIDGET(popover)))
-    gtk_popover_popdown(popover);
-  else
-    gtk_popover_popup(popover);
-}
-
 static void
 handle_chat_contacts_listbox_row_activated(GtkListBox *listbox,
                                            GtkListBoxRow *row,
@@ -159,22 +148,6 @@ handle_chat_messages_listbox_size_allocate(UNUSED GtkWidget *widget,
   handle->edge_value = upper - page_size;
 }
 
-static void
-handle_back_button_click(UNUSED GtkButton *button,
-			                   gpointer user_data)
-{
-  g_assert(user_data);
-
-  HdyLeaflet *leaflet = HDY_LEAFLET(user_data);
-
-  GList *children = gtk_container_get_children(GTK_CONTAINER(leaflet));
-
-  if (children) {
-    hdy_leaflet_set_visible_child(leaflet, GTK_WIDGET(children->data));
-    g_list_free(children);
-  }
-}
-
 static void
 handle_reveal_identity_button_click(GtkButton *button,
 				                            gpointer user_data)
@@ -247,25 +220,15 @@ handle_leave_chat_button_click(UNUSED GtkButton *button,
 
   UI_CHAT_Handle *handle = (UI_CHAT_Handle*) user_data;
 
-  if ((!handle) || (!(handle->send_text_view)))
-    return;
-
-  struct GNUNET_CHAT_Context *context = (struct GNUNET_CHAT_Context*) (
-    g_object_get_qdata(
-      G_OBJECT(handle->send_text_view),
-      handle->app->quarks.data
-    )
-  );
-
-  if (!context)
+  if ((!handle) || (!(handle->context)))
     return;
 
   struct GNUNET_CHAT_Contact *contact = GNUNET_CHAT_context_get_contact(
-    context
+    handle->context
   );
 
   struct GNUNET_CHAT_Group *group = GNUNET_CHAT_context_get_group(
-    context
+    handle->context
   );
 
   if (contact)
@@ -273,7 +236,9 @@ handle_leave_chat_button_click(UNUSED GtkButton *button,
   else if (group)
     GNUNET_CHAT_group_leave(group);
 
-  UI_CHAT_ENTRY_Handle *entry = GNUNET_CHAT_context_get_user_pointer(context);
+  UI_CHAT_ENTRY_Handle *entry = GNUNET_CHAT_context_get_user_pointer(
+    handle->context
+  );
 
   if ((!entry) || (!(handle->app)))
     return;
@@ -312,13 +277,105 @@ handle_chat_messages_sort(GtkListBoxRow* row0,
     return 0;
 }
 
+struct FilterTags
+{
+  const gchar *filter;
+  gboolean matching;
+};
+
+static enum GNUNET_GenericReturnValue
+_iterate_message_tags(void *cls,
+                      const struct GNUNET_CHAT_Message *message)
+{
+  g_assert((cls) && (message));
+
+  struct FilterTags *filterTags = (struct FilterTags*) cls;
+
+  const char *text = GNUNET_CHAT_message_get_text(message);
+  if (!text)
+    return GNUNET_YES;
+
+  gchar *_text = g_locale_to_utf8(text, -1, NULL, NULL, NULL);
+  if (!_text)
+    return GNUNET_YES;
+
+  if (g_strstr_len(_text, -1, filterTags->filter))
+    filterTags->matching = TRUE;
+
+  g_free(_text);
+  return filterTags->matching? GNUNET_NO : GNUNET_YES;
+}
+
+static gboolean
+handle_chat_messages_filter(GtkListBoxRow *row,
+                            gpointer user_data)
+{
+  g_assert((row) && (user_data));
+
+  MESSENGER_Application *app = (MESSENGER_Application*) user_data;
+
+  GtkListBox *listbox = GTK_LIST_BOX(gtk_widget_get_parent(GTK_WIDGET(row)));
+
+  if (!listbox)
+    return TRUE;
+
+  UI_CHAT_Handle *chat = (UI_CHAT_Handle*) (
+    g_object_get_qdata(G_OBJECT(listbox), app->quarks.ui)
+  );
+
+  if (!chat)
+    return TRUE;
+
+  const gchar *filter = gtk_entry_get_text(
+    GTK_ENTRY(chat->chat_search_entry)
+  );
+
+  if (!filter)
+    return TRUE;
+
+  UI_MESSAGE_Handle *message = (UI_MESSAGE_Handle*) (
+    g_object_get_qdata(G_OBJECT(row), app->quarks.ui)
+  );
+
+  if (!message)
+    return TRUE;
+
+  const gchar *sender = gtk_label_get_text(message->sender_label);
+  const gchar *text = gtk_label_get_text(message->text_label);
+
+  gboolean result = FALSE;
+
+  if (sender)
+    result |= g_str_match_string(filter, sender, TRUE);
+
+  if (text)
+    result |= g_str_match_string(filter, text, TRUE);
+
+  if (('#' == *filter) && (message->msg))
+  {
+    struct FilterTags filterTags;
+    filterTags.filter = &(filter[1]);
+    filterTags.matching = FALSE;
+
+    GNUNET_CHAT_message_iterate_tags(
+      message->msg,
+      _iterate_message_tags,
+      &filterTags
+    );
+
+    result |= filterTags.matching;
+  }
+
+  return result;
+}
+
 static void
 handle_chat_messages_selected_rows_changed(GtkListBox *listbox,
 					                                 gpointer user_data)
 {
   g_assert((listbox) && (user_data));
 
-  UI_CHAT_Handle *handle = (UI_CHAT_Handle*) user_data;
+  UI_CHAT_TITLE_Handle *handle = (UI_CHAT_TITLE_Handle*) user_data;
 
   GList *selected = gtk_list_box_get_selected_rows(listbox);
   uint32_t count = 0;
@@ -338,93 +395,14 @@ handle_chat_messages_selected_rows_changed(GtkListBox *listbox,
   gtk_label_set_text(handle->selection_count_label, counter->str);
   g_string_free(counter, TRUE);
 
+  gtk_widget_set_sensitive(GTK_WIDGET(handle->selection_tag_button), count == 1);
+
   if (count > 0)
     gtk_stack_set_visible_child(handle->chat_title_stack, handle->selection_box);
   else
     gtk_stack_set_visible_child(handle->chat_title_stack, handle->title_box);
 }
 
-static void
-handle_chat_selection_close_button_click(UNUSED GtkButton *button,
-					                               gpointer user_data)
-{
-  g_assert(user_data);
-
-  GtkListBox *listbox = GTK_LIST_BOX(user_data);
-
-  gtk_list_box_unselect_all(listbox);
-}
-
-void
-_delete_messages_callback(MESSENGER_Application *app,
-                          GList *selected,
-                          gulong delay)
-{
-  g_assert(app);
-
-  UI_MESSAGE_Handle *message;
-
-  while (selected)
-  {
-    GtkListBoxRow *row = GTK_LIST_BOX_ROW(selected->data);
-
-    if (!row)
-      goto skip_row;
-
-    message = (UI_MESSAGE_Handle*) g_object_get_qdata(
-      G_OBJECT(row),
-      app->quarks.ui
-    );
-
-    if ((!message) || (!(message->msg)))
-      goto skip_row;
-
-    GNUNET_CHAT_message_delete(
-    	message->msg,
-    	GNUNET_TIME_relative_multiply(
-    	  GNUNET_TIME_relative_get_second_(),
-	      delay
-	    )
-    );
-
-  skip_row:
-    selected = selected->next;
-  }
-}
-
-static void
-handle_chat_selection_delete_button_click(UNUSED GtkButton *button,
-					                                gpointer user_data)
-{
-  g_assert(user_data);
-
-  UI_CHAT_Handle *handle = (UI_CHAT_Handle*) user_data;
-
-  MESSENGER_Application *app = handle->app;
-
-  GList *selected = gtk_list_box_get_selected_rows(handle->messages_listbox);
-
-  if (app->settings.hide_delete_dialog)
-  {
-    _delete_messages_callback(app, selected, 0);
-    
-    if (selected)
-      g_list_free(selected);
-  }
-  else
-  {
-    ui_delete_messages_dialog_init(app, &(app->ui.delete_messages));
-
-    ui_delete_messages_dialog_link(
-      &(app->ui.delete_messages),
-      _delete_messages_callback,
-      selected
-    );
-
-    gtk_widget_show(GTK_WIDGET(app->ui.delete_messages.dialog));
-  }
-}
-
 static void
 handle_attach_file_button_click(GtkButton *button,
 				                        gpointer user_data)
@@ -545,12 +523,8 @@ _send_text_from_view(MESSENGER_Application *app,
     return FALSE;
   }
 
-  struct GNUNET_CHAT_Context *context = (struct GNUNET_CHAT_Context*) (
-    g_object_get_qdata(G_OBJECT(text_view), app->quarks.data)
-  );
-
-  if (context)
-    GNUNET_CHAT_context_send_text(context, text);
+  if (handle->context)
+    GNUNET_CHAT_context_send_text(handle->context, text);
 
   g_free(text);
   gtk_text_buffer_delete(buffer, &start, &end);
@@ -585,7 +559,7 @@ _drop_any_recording(UI_CHAT_Handle *handle)
 
 static void
 handle_sending_recording_upload_file(UNUSED void *cls,
-                                     const struct GNUNET_CHAT_File *file,
+                                     struct GNUNET_CHAT_File *file,
                                      uint64_t completed,
                                      uint64_t size)
 {
@@ -600,8 +574,8 @@ handle_sending_recording_upload_file(UNUSED void *cls,
 
   file_update_upload_info(file, completed, size);
 
-  if ((completed >= size) && (file_load->chat))
-    ui_chat_remove_file_load(file_load->chat, file_load);
+  if ((completed >= size) && (file_load->chat_title))
+    ui_chat_title_remove_file_load(file_load->chat_title, file_load);
 }
 
 static void
@@ -616,11 +590,7 @@ handle_send_record_button_click(GtkButton *button,
     g_object_get_qdata(G_OBJECT(button), app->quarks.ui)
   );
 
-  struct GNUNET_CHAT_Context *context = (struct GNUNET_CHAT_Context*) (
-    g_object_get_qdata(G_OBJECT(handle->send_text_view), app->quarks.data)
-  );
-
-  if ((handle->recorded) && (context) &&
+  if ((handle->recorded) && (handle->context) &&
       (handle->recording_filename[0]) &&
       (!gtk_revealer_get_child_revealed(handle->picker_revealer)))
   {
@@ -630,7 +600,7 @@ handle_send_record_button_click(GtkButton *button,
     gtk_progress_bar_set_fraction(file_load->load_progress_bar, 0.0);
 
     struct GNUNET_CHAT_File *file = GNUNET_CHAT_context_send_file(
-      context,
+      handle->context,
       handle->recording_filename,
       handle_sending_recording_upload_file,
       file_load
@@ -640,7 +610,7 @@ handle_send_record_button_click(GtkButton *button,
     {
       file_create_info(file);
 
-      ui_chat_add_file_load(handle, file_load);
+      ui_chat_title_add_file_load(handle->title, file_load);
     }
     else if (file_load)
       ui_file_load_entry_delete(file_load);
@@ -709,13 +679,17 @@ handle_send_now_button_click(GtkButton *button,
 
 static gboolean
 handle_send_record_button_pressed(GtkWidget *widget,
-                                  UNUSED GdkEvent *event,
+                                  GdkEvent *event,
                                   gpointer user_data)
 {
-  g_assert((widget) && (user_data));
+  g_assert((widget) && (event) && (user_data));
 
+  GdkEventButton *ev = (GdkEventButton*) event;
   MESSENGER_Application *app = (MESSENGER_Application*) user_data;
 
+  if (1 != ev->button)
+    return FALSE;
+
   GtkTextView *text_view = GTK_TEXT_VIEW(
     g_object_get_qdata(G_OBJECT(widget), app->quarks.widget)
   );
@@ -793,21 +767,33 @@ handle_send_record_button_pressed(GtkWidget *widget,
 
 static gboolean
 handle_send_record_button_released(GtkWidget *widget,
-                                   UNUSED GdkEvent *event,
+                                   GdkEvent *event,
                                    gpointer user_data)
 {
-  g_assert((widget) && (user_data));
+  g_assert((widget) && (event) && (user_data));
 
+  GdkEventButton *ev = (GdkEventButton*) event;
   MESSENGER_Application *app = (MESSENGER_Application*) user_data;
 
-  GtkTextView *text_view = GTK_TEXT_VIEW(
-    g_object_get_qdata(G_OBJECT(widget), app->quarks.widget)
-  );
-
   UI_CHAT_Handle *handle = (UI_CHAT_Handle*) (
     g_object_get_qdata(G_OBJECT(widget), app->quarks.ui)
   );
 
+  if ((gtk_stack_get_visible_child(handle->send_stack) == handle->send_text_box) &&
+      (3 == ev->button))
+  {
+    handle->send_pressed_time = UI_CHAT_SEND_BUTTON_HOLD_INTERVAL;
+
+    handle_send_record_button_click(GTK_BUTTON(widget), user_data);
+    return FALSE;
+  }
+  else if (1 != ev->button)
+    return FALSE;
+
+  GtkTextView *text_view = GTK_TEXT_VIEW(
+    g_object_get_qdata(G_OBJECT(widget), app->quarks.widget)
+  );
+
   handle->send_pressed_time = g_get_monotonic_time() - handle->send_pressed_time;
 
   GtkTextBuffer *buffer = gtk_text_view_get_buffer(text_view);
@@ -854,8 +840,7 @@ handle_send_text_key_press (GtkWidget *widget,
 
   MESSENGER_Application *app = (MESSENGER_Application*) user_data;
 
-  if ((app->settings.mobile_design) ||
-      (event->state & GDK_SHIFT_MASK) ||
+  if ((event->state & GDK_SHIFT_MASK) ||
       ((event->keyval != GDK_KEY_Return) &&
        (event->keyval != GDK_KEY_KP_Enter)))
     return FALSE;
@@ -900,7 +885,7 @@ _stop_playing_recording(UI_CHAT_Handle *handle,
 
   if (handle->play_timer)
   {
-    g_source_remove(handle->play_timer);
+    util_source_remove(handle->play_timer);
     handle->play_timer = 0;
   }
 }
@@ -920,13 +905,18 @@ handle_recording_play_button_click(UNUSED GtkButton *button,
     _stop_playing_recording(handle, TRUE);
   else if (handle->recording_filename[0])
   {
+    GString* uri = g_string_new("file://");
+    g_string_append(uri, handle->recording_filename);
+
     g_object_set(
-      G_OBJECT(handle->play_source),
-      "location",
-      handle->recording_filename,
+      G_OBJECT(handle->play_pipeline),
+      "uri",
+      uri->str,
       NULL
     );
 
+    g_string_free(uri, TRUE);
+
     gst_element_set_state(handle->play_pipeline, GST_STATE_PLAYING);
     handle->playing = TRUE;
 
@@ -938,6 +928,17 @@ handle_recording_play_button_click(UNUSED GtkButton *button,
   }
 }
 
+static void
+handle_search_entry_search_changed(UNUSED GtkSearchEntry* search_entry,
+                                   gpointer user_data)
+{
+  g_assert(user_data);
+
+  GtkListBox *listbox = GTK_LIST_BOX(user_data);
+
+  gtk_list_box_invalidate_filter(listbox);
+}
+
 static void
 handle_picker_button_click(UNUSED GtkButton *button,
 			                     gpointer user_data)
@@ -980,7 +981,7 @@ _record_timer_func(gpointer user_data)
   if (!(handle->recorded))
   {
     handle->record_time++;
-    handle->record_timer = g_timeout_add_seconds(
+    handle->record_timer = util_timeout_add_seconds(
       1,
       _record_timer_func,
       handle
@@ -998,11 +999,23 @@ _play_timer_func(gpointer user_data)
   g_assert(user_data);
 
   UI_CHAT_Handle *handle = (UI_CHAT_Handle*) user_data;
+  gint64 pos, len;
 
-  if (handle->play_time < handle->record_time * 100)
+  handle->play_timer = 0;
+
+  if (!(handle->play_pipeline))
+    return FALSE;
+
+  if (!gst_element_query_position(handle->play_pipeline, GST_FORMAT_TIME, &pos))
+    return FALSE;
+
+  if (!gst_element_query_duration(handle->play_pipeline, GST_FORMAT_TIME, &len))
+    return FALSE;
+
+  if (pos < len)
     gtk_progress_bar_set_fraction(
       handle->recording_progress_bar,
-      0.01 * handle->play_time / handle->record_time
+      1.0 * pos / len
     );
   else
     gtk_progress_bar_set_fraction(
@@ -1011,16 +1024,11 @@ _play_timer_func(gpointer user_data)
     );
 
   if (handle->playing)
-  {
-    handle->play_time++;
-    handle->play_timer = g_timeout_add(
+    handle->play_timer = util_timeout_add(
       10,
       _play_timer_func,
       handle
     );
-  }
-  else
-    handle->play_timer = 0;
 
   return FALSE;
 }
@@ -1039,8 +1047,7 @@ handle_record_bus_watch(UNUSED GstBus *bus,
   {
     case GST_MESSAGE_STREAM_START:
       handle->record_time = 0;
-      handle->record_timer = g_timeout_add_seconds(
-        0,
+      handle->record_timer = util_idle_add(
         _record_timer_func,
         handle
       );
@@ -1063,15 +1070,20 @@ handle_play_bus_watch(UNUSED GstBus *bus,
 
   switch (type)
   {
-    case GST_MESSAGE_STREAM_START:
-      handle->play_time = 0;
-      handle->play_timer = g_timeout_add_seconds(
-        0,
-        _play_timer_func,
-        handle
-      );
+    case GST_MESSAGE_STATE_CHANGED:
+    {
+      GstState old_state, new_state, pending_state;
+      gst_message_parse_state_changed(msg, &old_state, &new_state, &pending_state);
 
+      if (GST_STATE_PLAYING == new_state)
+        handle->play_timer = util_idle_add(
+          _play_timer_func,
+          handle
+        );
+      else if (GST_STATE_PLAYING == old_state)
+        _stop_playing_recording(handle, FALSE);
       break;
+    }
     case GST_MESSAGE_EOS:
       if (handle->playing)
 	      _stop_playing_recording(handle, FALSE);
@@ -1109,13 +1121,17 @@ _setup_gst_pipelines(UI_CHAT_Handle *handle)
     gst_object_unref(bus);
   }
 
-  handle->play_pipeline = gst_parse_launch(
-    "filesrc name=source ! oggdemux ! vorbisdec ! audioconvert ! autoaudiosink",
-    NULL
-  );
+  handle->play_pipeline = gst_element_factory_make("playbin", NULL);
+  handle->play_sink = gst_element_factory_make("autoaudiosink", "asink");
 
-  handle->play_source = gst_bin_get_by_name(
-    GST_BIN(handle->play_pipeline), "source"
+  if ((!(handle->play_pipeline)) || (!(handle->play_sink)))
+    return;
+
+  g_object_set(
+      G_OBJECT(handle->play_pipeline),
+      "audio-sink",
+      handle->play_sink,
+      NULL
   );
 
   {
@@ -1132,22 +1148,23 @@ _setup_gst_pipelines(UI_CHAT_Handle *handle)
 }
 
 UI_CHAT_Handle*
-ui_chat_new(MESSENGER_Application *app)
+ui_chat_new(MESSENGER_Application *app,
+            struct GNUNET_CHAT_Context *context)
 {
-  g_assert(app);
+  g_assert((app) && (context));
 
   UI_CHAT_Handle *handle = g_malloc(sizeof(UI_CHAT_Handle));
-  UI_MESSENGER_Handle *messenger = &(app->ui.messenger);
 
   memset(handle, 0, sizeof(*handle));
 
   _setup_gst_pipelines(handle);
 
   handle->app = app;
+  handle->context = context;
 
-  handle->loads = NULL;
+  handle->title = ui_chat_title_new(handle->app, handle);
 
-  handle->builder = gtk_builder_new_from_resource(
+  handle->builder = ui_builder_from_resource(
     application_get_resource_path(app, "ui/chat.ui")
   );
 
@@ -1155,81 +1172,16 @@ ui_chat_new(MESSENGER_Application *app)
     gtk_builder_get_object(handle->builder, "chat_box")
   );
 
-  handle->back_button = GTK_BUTTON(
-    gtk_builder_get_object(handle->builder, "back_button")
-  );
-
-  g_object_bind_property(
-    messenger->leaflet_chat,
-    "folded",
-    handle->back_button,
-    "visible",
-    G_BINDING_SYNC_CREATE
-  );
-
-  g_signal_connect(
-    handle->back_button,
-    "clicked",
-    G_CALLBACK(handle_back_button_click),
-    messenger->leaflet_chat
-  );
-
   handle->flap_chat_details = HDY_FLAP(
     gtk_builder_get_object(handle->builder, "flap_chat_details")
   );
 
-  handle->chat_title_stack = GTK_STACK(
-    gtk_builder_get_object(handle->builder, "chat_title_stack")
+  handle->chat_search_bar = HDY_SEARCH_BAR(
+    gtk_builder_get_object(handle->builder, "chat_search_bar")
   );
 
-  handle->title_box = GTK_WIDGET(
-    gtk_builder_get_object(handle->builder, "title_box")
-  );
-
-  handle->selection_box = GTK_WIDGET(
-    gtk_builder_get_object(handle->builder, "selection_box")
-  );
-
-  handle->chat_avatar = HDY_AVATAR(
-    gtk_builder_get_object(handle->builder, "chat_avatar")
-  );
-
-  handle->chat_title = GTK_LABEL(
-    gtk_builder_get_object(handle->builder, "chat_title")
-  );
-
-  handle->chat_subtitle = GTK_LABEL(
-    gtk_builder_get_object(handle->builder, "chat_subtitle")
-  );
-
-  handle->chat_load_button = GTK_BUTTON(
-    gtk_builder_get_object(handle->builder, "chat_load_button")
-  );
-
-  handle->chat_load_popover = GTK_POPOVER(
-    gtk_builder_get_object(handle->builder, "chat_load_popover")
-  );
-
-  handle->chat_load_listbox = GTK_LIST_BOX(
-    gtk_builder_get_object(handle->builder, "chat_load_listbox")
-  );
-
-  g_signal_connect(
-    handle->chat_load_button,
-    "clicked",
-    G_CALLBACK(handle_popover_via_button_click),
-    handle->chat_load_popover
-  );
-
-  handle->chat_details_button = GTK_BUTTON(
-    gtk_builder_get_object(handle->builder, "chat_details_button")
-  );
-
-  g_signal_connect(
-    handle->chat_details_button,
-    "clicked",
-    G_CALLBACK(handle_chat_details_via_button_click),
-    handle
+  handle->chat_search_entry = GTK_SEARCH_ENTRY(
+    gtk_builder_get_object(handle->builder, "chat_search_entry")
   );
 
   handle->chat_details_label = GTK_LABEL(
@@ -1255,6 +1207,10 @@ ui_chat_new(MESSENGER_Application *app)
     gtk_builder_get_object(handle->builder, "chat_details_files_box")
   );
 
+  handle->chat_details_media_box = GTK_BOX(
+    gtk_builder_get_object(handle->builder, "chat_details_media_box")
+  );
+
   handle->chat_details_avatar = HDY_AVATAR(
     gtk_builder_get_object(handle->builder, "chat_details_avatar")
   );
@@ -1311,18 +1267,6 @@ ui_chat_new(MESSENGER_Application *app)
     gtk_builder_get_object(handle->builder, "chat_notifications_switch")
   );
 
-  handle->selection_close_button = GTK_BUTTON(
-    gtk_builder_get_object(handle->builder, "selection_close_button")
-  );
-
-  handle->selection_count_label = GTK_LABEL(
-    gtk_builder_get_object(handle->builder, "selection_count_label")
-  );
-
-  handle->selection_delete_button = GTK_BUTTON(
-    gtk_builder_get_object(handle->builder, "selection_delete_button")
-  );
-
   handle->chat_scrolled_window = GTK_SCROLLED_WINDOW(
     gtk_builder_get_object(handle->builder, "chat_scrolled_window")
   );
@@ -1342,6 +1286,10 @@ ui_chat_new(MESSENGER_Application *app)
     gtk_builder_get_object(handle->builder, "chat_files_listbox")
   );
 
+  handle->chat_media_flowbox = GTK_FLOW_BOX(
+    gtk_builder_get_object(handle->builder, "chat_media_flowbox")
+  );
+
   handle->messages_listbox = GTK_LIST_BOX(
     gtk_builder_get_object(handle->builder, "messages_listbox")
   );
@@ -1353,25 +1301,25 @@ ui_chat_new(MESSENGER_Application *app)
     NULL
   );
 
-  g_signal_connect(
+  gtk_list_box_set_filter_func(
     handle->messages_listbox,
-    "selected-rows-changed",
-    G_CALLBACK(handle_chat_messages_selected_rows_changed),
-    handle
+    handle_chat_messages_filter,
+    app,
+    NULL
   );
 
   g_signal_connect(
-    handle->selection_close_button,
-    "clicked",
-    G_CALLBACK(handle_chat_selection_close_button_click),
+    handle->chat_search_entry,
+    "search-changed",
+    G_CALLBACK(handle_search_entry_search_changed),
     handle->messages_listbox
   );
 
   g_signal_connect(
-    handle->selection_delete_button,
-    "clicked",
-    G_CALLBACK(handle_chat_selection_delete_button_click),
-    handle
+    handle->messages_listbox,
+    "selected-rows-changed",
+    G_CALLBACK(handle_chat_messages_selected_rows_changed),
+    handle->title
   );
 
   g_signal_connect(
@@ -1527,6 +1475,12 @@ ui_chat_new(MESSENGER_Application *app)
     handle
   );
 
+  g_object_set_qdata(
+    G_OBJECT(handle->messages_listbox),
+    app->quarks.ui,
+    handle
+  );
+
   handle->recording_close_button = GTK_BUTTON(
     gtk_builder_get_object(handle->builder, "recording_close_button")
   );
@@ -1582,21 +1536,21 @@ ui_chat_new(MESSENGER_Application *app)
   return handle;
 }
 
-struct IterateChatGroupClosure {
+struct IterateChatClosure {
   MESSENGER_Application *app;
-  GtkListBox *listbox;
+  GtkContainer *container;
 };
 
-static int
+static enum GNUNET_GenericReturnValue
 iterate_ui_chat_update_group_contacts(void *cls,
                                       UNUSED const struct GNUNET_CHAT_Group *group,
                                       struct GNUNET_CHAT_Contact *contact)
 {
-  struct IterateChatGroupClosure *closure = (
-    (struct IterateChatGroupClosure*) cls
+  struct IterateChatClosure *closure = (
+    (struct IterateChatClosure*) cls
   );
 
-  GtkListBox *listbox = closure->listbox;
+  GtkListBox *listbox = GTK_LIST_BOX(closure->container);
   UI_ACCOUNT_ENTRY_Handle* entry = ui_account_entry_new(closure->app);
 
   ui_account_entry_set_contact(entry, contact);
@@ -1618,58 +1572,12 @@ iterate_ui_chat_update_group_contacts(void *cls,
   return GNUNET_YES;
 }
 
-void
-ui_chat_update(UI_CHAT_Handle *handle,
-               MESSENGER_Application *app,
-               struct GNUNET_CHAT_Context* context)
+static void
+_chat_update_contacts(UI_CHAT_Handle *handle,
+                      MESSENGER_Application *app,
+                      struct GNUNET_CHAT_Group* group)
 {
-  g_assert((handle) && (app) && (context));
-
-  struct GNUNET_CHAT_Contact* contact;
-  struct GNUNET_CHAT_Group* group;
-
-  contact = GNUNET_CHAT_context_get_contact(context);
-  group = GNUNET_CHAT_context_get_group(context);
-
-  const char *title = NULL;
-  const char *icon = "action-unavailable-symbolic";
-
-  GString *subtitle = g_string_new("");
-
-  if (contact)
-  {
-    title = GNUNET_CHAT_contact_get_name(contact);
-    icon = "avatar-default-symbolic";
-  }
-  else if (group)
-  {
-    title = GNUNET_CHAT_group_get_name(group);
-
-    if ((title) && ('#' == *title))
-      icon = "network-wired-symbolic";
-    else
-      icon = "system-users-symbolic";
-
-    g_string_append_printf(
-      subtitle,
-      _("%d members"),
-      GNUNET_CHAT_group_iterate_contacts(group, NULL, NULL)
-    );
-  }
-
-  ui_avatar_set_text(handle->chat_avatar, title);
-  hdy_avatar_set_icon_name(handle->chat_avatar, icon);
-
-  ui_avatar_set_text(handle->chat_details_avatar, title);
-  hdy_avatar_set_icon_name(handle->chat_details_avatar, icon);
-
-  ui_label_set_text(handle->chat_title, title);
-  ui_label_set_text(handle->chat_details_label, title);
-
-  if (subtitle->len > 0)
-    gtk_label_set_text(handle->chat_subtitle, subtitle->str);
-
-  g_string_free(subtitle, TRUE);
+  g_assert((handle) && (app));
 
   GList* children = gtk_container_get_children(
     GTK_CONTAINER(handle->chat_contacts_listbox)
@@ -1691,9 +1599,9 @@ ui_chat_update(UI_CHAT_Handle *handle,
 
   if (group)
   {
-    struct IterateChatGroupClosure closure;
+    struct IterateChatClosure closure;
     closure.app = app;
-    closure.listbox = handle->chat_contacts_listbox;
+    closure.container = GTK_CONTAINER(handle->chat_contacts_listbox);
 
     GNUNET_CHAT_group_iterate_contacts(
 	    group,
@@ -1706,6 +1614,222 @@ ui_chat_update(UI_CHAT_Handle *handle,
     GTK_WIDGET(handle->chat_details_contacts_box),
     group? TRUE : FALSE
   );
+}
+
+static enum GNUNET_GenericReturnValue
+iterate_ui_chat_update_context_files(void *cls,
+                                     struct GNUNET_CHAT_Context *context,
+                                     struct GNUNET_CHAT_File *file)
+{
+  struct IterateChatClosure *closure = (
+    (struct IterateChatClosure*) cls
+  );
+
+  GtkListBox *listbox = GTK_LIST_BOX(closure->container);
+  UI_FILE_ENTRY_Handle* entry = ui_file_entry_new(closure->app);
+  ui_file_entry_update(entry, file);
+
+  gtk_list_box_prepend(listbox, entry->entry_box);
+
+  GtkListBoxRow *row = GTK_LIST_BOX_ROW(
+    gtk_widget_get_parent(entry->entry_box)
+  );
+
+  g_object_set_qdata(G_OBJECT(row), closure->app->quarks.data, file);
+  g_object_set_qdata_full(
+    G_OBJECT(row),
+    closure->app->quarks.ui,
+    entry,
+    (GDestroyNotify) ui_file_entry_delete
+  );
+
+  return GNUNET_YES;
+}
+
+static void
+_chat_update_files(UI_CHAT_Handle *handle,
+                   MESSENGER_Application *app,
+                   struct GNUNET_CHAT_Context *context)
+{
+  g_assert((handle) && (app));
+
+  GList* children = gtk_container_get_children(
+    GTK_CONTAINER(handle->chat_files_listbox)
+  );
+
+  GList *item = children;
+  while (item) {
+    GtkWidget *widget = GTK_WIDGET(item->data);
+    item = item->next;
+
+    gtk_container_remove(
+      GTK_CONTAINER(handle->chat_files_listbox),
+      widget
+    );
+  }
+
+  if (children)
+    g_list_free(children);
+
+  struct IterateChatClosure closure;
+  closure.app = app;
+  closure.container = GTK_CONTAINER(handle->chat_files_listbox);
+
+  const int count = context? GNUNET_CHAT_context_iterate_files(
+    context,
+    iterate_ui_chat_update_context_files,
+    &closure
+  ) : 0;
+
+  gtk_widget_set_visible(
+    GTK_WIDGET(handle->chat_details_files_box),
+    count? TRUE : FALSE
+  );
+}
+
+static enum GNUNET_GenericReturnValue
+iterate_ui_chat_update_context_media(void *cls,
+                                     struct GNUNET_CHAT_Context *context,
+                                     struct GNUNET_CHAT_File *file)
+{
+  struct IterateChatClosure *closure = (
+    (struct IterateChatClosure*) cls
+  );
+
+  GtkFlowBox *flowbox = GTK_FLOW_BOX(closure->container);
+  UI_MEDIA_PREVIEW_Handle* handle = ui_media_preview_new(closure->app);
+  ui_media_preview_update(handle, file);
+
+  GdkPixbuf *image = file_get_current_preview_image(file);
+
+  if (!image)
+  {
+    ui_media_preview_delete(handle);
+    return GNUNET_YES;
+  }
+
+  gtk_flow_box_insert(flowbox, handle->media_box, 0);
+
+  GtkFlowBoxChild *child = GTK_FLOW_BOX_CHILD(
+    gtk_widget_get_parent(handle->media_box)
+  );
+
+  g_object_set_qdata(G_OBJECT(child), closure->app->quarks.data, file);
+  g_object_set_qdata(
+    G_OBJECT(child),
+    closure->app->quarks.ui,
+    handle
+  );
+
+  gtk_widget_set_size_request(GTK_WIDGET(child), 80, 80);
+
+  gtk_widget_show_all(GTK_WIDGET(child));
+  return GNUNET_YES;
+}
+
+static void
+_chat_update_media(UI_CHAT_Handle *handle,
+                   MESSENGER_Application *app,
+                   struct GNUNET_CHAT_Context *context)
+{
+  g_assert((handle) && (app));
+
+  GList* children = gtk_container_get_children(
+    GTK_CONTAINER(handle->chat_media_flowbox)
+  );
+
+  GList *item = children;
+  while (item) {
+    GtkWidget *widget = GTK_WIDGET(item->data);
+    item = item->next;
+
+    UI_MEDIA_PREVIEW_Handle *media = g_object_get_qdata(
+      G_OBJECT(widget),
+      app->quarks.ui
+    );
+
+    ui_media_preview_delete(media);
+
+    gtk_container_remove(
+      GTK_CONTAINER(handle->chat_media_flowbox),
+      widget
+    );
+  }
+
+  if (children)
+    g_list_free(children);
+
+  struct IterateChatClosure closure;
+  closure.app = app;
+  closure.container = GTK_CONTAINER(handle->chat_media_flowbox);
+
+  const int count = context? GNUNET_CHAT_context_iterate_files(
+    context,
+    iterate_ui_chat_update_context_media,
+    &closure
+  ) : 0;
+
+  gtk_widget_set_visible(
+    GTK_WIDGET(handle->chat_details_media_box),
+    count? TRUE : FALSE
+  );
+}
+
+static const gchar*
+_chat_get_default_subtitle(UI_CHAT_Handle *handle,
+                           MESSENGER_Application *app,
+                           struct GNUNET_CHAT_Group* group)
+{
+  g_assert((handle) && (app));
+
+  GList *rows = gtk_container_get_children(
+    GTK_CONTAINER(handle->messages_listbox)
+  );
+
+  if (!rows)
+    return NULL;
+
+  UI_MESSAGE_Handle *last_message = NULL;
+  for (GList *row = rows; row; row = row->next)
+  {
+    UI_MESSAGE_Handle *message = (UI_MESSAGE_Handle*) g_object_get_qdata(
+      G_OBJECT(row->data), app->quarks.ui
+    );
+
+    if (!message)
+      continue;
+
+    ui_message_refresh(message);
+    last_message = message;
+  }
+
+  g_list_free(rows);
+
+  if ((!last_message) || (!(last_message->timestamp_label)))
+    return NULL;
+
+  return group? NULL : gtk_label_get_text(last_message->timestamp_label);
+}
+
+void
+ui_chat_update(UI_CHAT_Handle *handle,
+               MESSENGER_Application *app)
+{
+  g_assert((handle) && (app));
+
+  struct GNUNET_CHAT_Contact* contact;
+  struct GNUNET_CHAT_Group* group;
+
+  contact = GNUNET_CHAT_context_get_contact(handle->context);
+  group = GNUNET_CHAT_context_get_group(handle->context);
+
+  _chat_update_contacts(handle, app, group);
+  _chat_update_files(handle, app, handle->context);
+  _chat_update_media(handle, app, handle->context);
+
+  const gchar *subtitle = _chat_get_default_subtitle(handle, app, group);
+
+  ui_chat_title_update(handle->title, app, subtitle);
 
   g_object_set_qdata(
     G_OBJECT(handle->reveal_identity_button),
@@ -1739,7 +1863,7 @@ ui_chat_update(UI_CHAT_Handle *handle,
     (contact) || (group)? TRUE : FALSE
   );
 
-  const int status = GNUNET_CHAT_context_get_status(context);
+  const int status = GNUNET_CHAT_context_get_status(handle->context);
   const gboolean activated = (GNUNET_OK == status? TRUE : FALSE);
 
   gtk_text_view_set_editable(handle->send_text_view, activated);
@@ -1748,37 +1872,6 @@ ui_chat_update(UI_CHAT_Handle *handle,
   gtk_widget_set_sensitive(GTK_WIDGET(handle->attach_file_button), activated);
   gtk_widget_set_sensitive(GTK_WIDGET(handle->emoji_button), activated);
   gtk_widget_set_sensitive(GTK_WIDGET(handle->send_record_button), activated);
-
-  GList *rows = gtk_container_get_children(
-    GTK_CONTAINER(handle->messages_listbox)
-  );
-
-  if (!rows)
-    return;
-
-  UI_MESSAGE_Handle *last_message = NULL;
-  for (GList *row = rows; row; row = row->next)
-  {
-    UI_MESSAGE_Handle *message = (UI_MESSAGE_Handle*) g_object_get_qdata(
-      G_OBJECT(row->data), app->quarks.ui
-    );
-
-    if (!message)
-      continue;
-
-    ui_message_refresh(message);
-    last_message = message;
-  }
-
-  g_list_free(rows);
-
-  if ((!last_message) || (!(last_message->timestamp_label)))
-    return;
-
-  const gchar *time = gtk_label_get_text(last_message->timestamp_label);
-
-  if (!group)
-    gtk_label_set_text(handle->chat_subtitle, time);
 }
 
 void
@@ -1786,12 +1879,39 @@ ui_chat_delete(UI_CHAT_Handle *handle)
 {
   g_assert(handle);
 
+  GList *message_rows = gtk_container_get_children(GTK_CONTAINER(handle->messages_listbox));
+  GList *row_element = message_rows;
+
+  while (row_element)
+  {
+    GtkWidget *row = GTK_WIDGET(row_element->data);
+
+    if (!row)
+      goto skip_row;
+
+    UI_MESSAGE_Handle *message = (UI_MESSAGE_Handle*) g_object_get_qdata(
+      G_OBJECT(row), handle->app->quarks.ui
+    );
+
+    if (message)
+      ui_chat_remove_message(handle, handle->app, message);
+
+  skip_row:
+    row_element = row_element->next;
+  }
+
+  if (message_rows)
+    g_list_free(message_rows);
+
   ui_picker_delete(handle->picker);
 
-  g_object_unref(handle->builder);
+  _chat_update_contacts(handle, handle->app, NULL);
+  _chat_update_media(handle, handle->app, NULL);
+  _chat_update_files(handle, handle->app, NULL);
 
-  if (handle->loads)
-    g_list_free_full(handle->loads, (GDestroyNotify) ui_file_load_entry_delete);
+  ui_chat_title_delete(handle->title);
+
+  g_object_unref(handle->builder);
 
   if (handle->record_pipeline)
   {
@@ -1809,10 +1929,10 @@ ui_chat_delete(UI_CHAT_Handle *handle)
     remove(handle->recording_filename);
 
   if (handle->record_timer)
-    g_source_remove(handle->record_timer);
+    util_source_remove(handle->record_timer);
 
   if (handle->play_timer)
-    g_source_remove(handle->play_timer);
+    util_source_remove(handle->play_timer);
 
   g_free(handle);
 }
@@ -1837,7 +1957,7 @@ ui_chat_add_message(UI_CHAT_Handle *handle,
 
 void
 ui_chat_remove_message(UI_CHAT_Handle *handle,
-                       UNUSED MESSENGER_Application *app,
+                       MESSENGER_Application *app,
                        UI_MESSAGE_Handle *message)
 {
   g_assert((handle) && (message) && (message->message_box));
@@ -1850,48 +1970,5 @@ ui_chat_remove_message(UI_CHAT_Handle *handle,
   if (parent == GTK_WIDGET(handle->messages_listbox))
     gtk_container_remove(GTK_CONTAINER(handle->messages_listbox), row);
 
-  ui_message_delete(message);
-}
-
-void
-ui_chat_add_file_load(UI_CHAT_Handle *handle,
-		      UI_FILE_LOAD_ENTRY_Handle *file_load)
-{
-  g_assert((handle) && (file_load));
-
-  gtk_container_add(
-      GTK_CONTAINER(handle->chat_load_listbox),
-      file_load->entry_box
-  );
-
-  handle->loads = g_list_append(handle->loads, file_load);
-
-  gtk_widget_show(GTK_WIDGET(handle->chat_load_button));
-
-  file_load->chat = handle;
-}
-
-void
-ui_chat_remove_file_load(UI_CHAT_Handle *handle,
-			 UI_FILE_LOAD_ENTRY_Handle *file_load)
-{
-  g_assert((handle) && (file_load) && (handle == file_load->chat) &&
-		(file_load->entry_box));
-
-  handle->loads = g_list_remove(handle->loads, file_load);
-
-  gtk_container_remove(
-      GTK_CONTAINER(handle->chat_load_listbox),
-      gtk_widget_get_parent(file_load->entry_box)
-  );
-
-  if (handle->loads)
-    return;
-
-  if (gtk_widget_is_visible(GTK_WIDGET(handle->chat_load_popover)))
-    gtk_popover_popdown(handle->chat_load_popover);
-
-  gtk_widget_hide(GTK_WIDGET(handle->chat_load_button));
-
-  file_load->chat = NULL;
+  ui_message_delete(message, app);
 }
