@@ -24,26 +24,31 @@
 
 #include "chat_entry.h"
 
+#include "chat_title.h"
 #include "message.h"
 
 #include "../application.h"
+#include "../contact.h"
 #include "../ui.h"
+
 #include <gnunet/gnunet_chat_lib.h>
 #include <gnunet/gnunet_time_lib.h>
 
 UI_CHAT_ENTRY_Handle*
-ui_chat_entry_new(MESSENGER_Application *app)
+ui_chat_entry_new(MESSENGER_Application *app,
+                  struct GNUNET_CHAT_Context *context)
 {
-  g_assert(app);
+  g_assert((app) && (context));
 
   UI_CHAT_ENTRY_Handle* handle = g_malloc(sizeof(UI_CHAT_ENTRY_Handle));
 
   memset(handle, 0, sizeof(*handle));
 
   handle->timestamp = GNUNET_TIME_absolute_get_zero_();
+  handle->context = context;
 
-  handle->chat = ui_chat_new(app);
-  handle->builder = gtk_builder_new_from_resource(
+  handle->chat = ui_chat_new(app, handle->context);
+  handle->builder = ui_builder_from_resource(
     application_get_resource_path(app, "ui/chat_entry.ui")
   );
 
@@ -71,48 +76,82 @@ ui_chat_entry_new(MESSENGER_Application *app)
     gtk_builder_get_object(handle->builder, "read_receipt_image")
   );
 
+  GNUNET_CHAT_context_set_user_pointer(
+    handle->context,
+    handle
+  );
+
   return handle;
 }
 
+static void
+_chat_entry_update_contact(UI_CHAT_ENTRY_Handle *handle,
+                           MESSENGER_Application *app,
+                           struct GNUNET_CHAT_Contact* contact)
+{
+  g_assert((handle) && (app));
+
+  struct GNUNET_CHAT_Contact *prev = g_object_get_qdata(
+    G_OBJECT(handle->entry_avatar),
+    app->quarks.data
+  );
+
+  if (prev)
+  {
+    contact_remove_name_label_from_info(contact, handle->title_label);
+    contact_remove_name_avatar_from_info(contact, handle->entry_avatar);
+  }
+  
+  if (contact)
+  {
+    contact_add_name_label_to_info(contact, handle->title_label);
+    contact_add_name_avatar_to_info(contact, handle->entry_avatar);
+  }
+
+  g_object_set_qdata(
+    G_OBJECT(handle->entry_avatar),
+    app->quarks.data,
+    contact
+  );
+}
+
 void
 ui_chat_entry_update(UI_CHAT_ENTRY_Handle *handle,
-		                 MESSENGER_Application *app,
-		                 struct GNUNET_CHAT_Context *context)
+		                 MESSENGER_Application *app)
 {
   g_assert((handle) && (app));
 
-  const struct GNUNET_CHAT_Contact* contact;
-  const struct GNUNET_CHAT_Group* group;
+  struct GNUNET_CHAT_Contact* contact;
+  struct GNUNET_CHAT_Group* group;
 
-  contact = GNUNET_CHAT_context_get_contact(context);
-  group = GNUNET_CHAT_context_get_group(context);
+  contact = GNUNET_CHAT_context_get_contact(handle->context);
+  group = GNUNET_CHAT_context_get_group(handle->context);
 
-  const char *title = NULL;
   const char *icon = "action-unavailable-symbolic";
 
+  _chat_entry_update_contact(handle, app, contact);
+
   if (contact)
-  {
-    title = GNUNET_CHAT_contact_get_name(contact);
     icon = "avatar-default-symbolic";
-  }
   else if (group)
   {
-    title = GNUNET_CHAT_group_get_name(group);
+    const char *title = GNUNET_CHAT_group_get_name(group);
 
     if ((title) && ('#' == *title))
       icon = "network-wired-symbolic";
     else
       icon = "system-users-symbolic";
+
+    ui_label_set_text(handle->title_label, title);
+    ui_avatar_set_text(handle->entry_avatar, title);
   }
 
-  ui_label_set_text(handle->title_label, title);
-  ui_avatar_set_text(handle->entry_avatar, title);
   hdy_avatar_set_icon_name(handle->entry_avatar, icon);
 
   if (!(handle->chat))
     return;
 
-  ui_chat_update(handle->chat, app, context);
+  ui_chat_update(handle->chat, app);
 
   GList *rows = gtk_container_get_children(
     GTK_CONTAINER(handle->chat->messages_listbox)
@@ -141,18 +180,40 @@ ui_chat_entry_update(UI_CHAT_ENTRY_Handle *handle,
 
   handle->timestamp = last_message->timestamp;
 
+  const struct GNUNET_TIME_Timestamp timestamp = GNUNET_TIME_absolute_to_timestamp(
+    handle->timestamp
+  );
+
+  GDateTime *dt_now = g_date_time_new_now_local();
+  GDateTime *dt_message = g_date_time_new_from_unix_local(
+    (gint64) (timestamp.abs_time.abs_value_us / 1000000)
+  );
+
+  GTimeSpan span = g_date_time_difference(dt_now, dt_message);
+  gchar *time = NULL;
+
+  if (span > 7 * G_TIME_SPAN_DAY)
+    time = g_date_time_format(dt_message, "%F");
+  else if (span > 2 * G_TIME_SPAN_DAY)
+    time = g_date_time_format(dt_message, "%A");
+  else if (span > G_TIME_SPAN_DAY)
+    time = g_date_time_format(dt_message, _("Yesterday"));
+  else
+    time = g_date_time_format(dt_message, "%R");
+
+  g_date_time_unref(dt_now);
+  g_date_time_unref(dt_message);
+
   const gchar *text = gtk_label_get_text(last_message->text_label);
-  const gchar *time = gtk_label_get_text(last_message->timestamp_label);
 
   if (group)
   {
-    GString *message_text = g_string_new(
-	    gtk_label_get_text(last_message->sender_label)
-    );
+    GString *message_text = g_string_new(NULL);
 
     g_string_append_printf(
       message_text,
-      _(": %s"),
+      "%s: %s",
+      gtk_label_get_text(last_message->sender_label),
       text
     );
 
@@ -162,7 +223,11 @@ ui_chat_entry_update(UI_CHAT_ENTRY_Handle *handle,
   else
     gtk_label_set_text(handle->text_label, text);
 
-  gtk_label_set_text(handle->timestamp_label, time);
+  if (time)
+  {
+    gtk_label_set_text(handle->timestamp_label, time);
+    g_free(time);
+  }
 
   gtk_widget_set_visible(
     GTK_WIDGET(handle->read_receipt_image),
@@ -184,7 +249,10 @@ ui_chat_entry_delete(UI_CHAT_ENTRY_Handle *handle)
   g_object_unref(handle->builder);
 
   if (handle->update)
-    g_source_remove(handle->update);
+    util_source_remove(handle->update);
+
+  if (handle->context)
+    GNUNET_CHAT_context_set_user_pointer(handle->context, NULL);
 
   g_free(handle);
 }
@@ -197,6 +265,8 @@ ui_chat_entry_dispose(UI_CHAT_ENTRY_Handle *handle,
 
   UI_MESSENGER_Handle *ui = &(app->ui.messenger);
 
+  util_source_remove_by_data(handle);
+
   ui->chat_entries = g_list_remove(ui->chat_entries, handle);
 
   gtk_container_remove(
@@ -204,15 +274,12 @@ ui_chat_entry_dispose(UI_CHAT_ENTRY_Handle *handle,
     gtk_widget_get_parent(handle->entry_box)
   );
 
-  struct GNUNET_CHAT_Context *context = (struct GNUNET_CHAT_Context*) (
-    g_object_get_qdata(
-      G_OBJECT(handle->chat->send_text_view),
-      app->quarks.data
-    )
-  );
+  _chat_entry_update_contact(handle, app, NULL);
 
-  if (context)
-    GNUNET_CHAT_context_set_user_pointer(context, NULL);
+  gtk_container_remove(
+    GTK_CONTAINER(ui->chat_title_stack),
+    handle->chat->title->chat_title_box
+  );
 
   gtk_container_remove(
     GTK_CONTAINER(ui->chats_stack),
