@@ -27,6 +27,8 @@
 #include <gnunet/gnunet_chat_lib.h>
 #include <gnunet/gnunet_common.h>
 
+#include "tag.h"
+
 #include "../application.h"
 #include "../contact.h"
 #include "../file.h"
@@ -34,7 +36,7 @@
 
 static void
 handle_downloading_file(void *cls,
-                        const struct GNUNET_CHAT_File *file,
+                        struct GNUNET_CHAT_File *file,
                         uint64_t completed,
                         uint64_t size)
 {
@@ -89,7 +91,11 @@ handle_file_button_click(GtkButton *button,
   }
   else if (local_size < size)
   {
-    GNUNET_CHAT_file_start_download(file, handle_downloading_file, app);
+    GNUNET_CHAT_file_start_download(
+      file,
+      handle_downloading_file,
+      app
+    );
 
     gtk_image_set_from_icon_name(
     	handle->file_status_image,
@@ -156,24 +162,6 @@ handle_media_button_click(GtkButton *button,
   g_string_free(uri, TRUE);
 }
 
-static int
-handle_message_redraw_animation(gpointer user_data)
-{
-  g_assert(user_data);
-
-  UI_MESSAGE_Handle *handle = (UI_MESSAGE_Handle*) user_data;
-
-  handle->redraw_animation = 0;
-
-  if ((handle->preview_drawing_area) &&
-      ((handle->preview_image) ||
-       (handle->preview_animation) ||
-       (handle->preview_animation_iter)))
-    gtk_widget_queue_draw(GTK_WIDGET(handle->preview_drawing_area));
-
-  return FALSE;
-}
-
 static gboolean
 handle_preview_drawing_area_draw(GtkWidget* drawing_area,
                                  cairo_t* cairo,
@@ -190,29 +178,16 @@ handle_preview_drawing_area_draw(GtkWidget* drawing_area,
 
   gtk_render_background(context, cairo, 0, 0, width, height);
 
-  GdkPixbuf *image = handle->preview_image;
-
-  if (!(handle->preview_animation))
-    goto render_image;
-
-  if (handle->preview_animation_iter)
-    gdk_pixbuf_animation_iter_advance(handle->preview_animation_iter, NULL);
-  else
-    handle->preview_animation_iter = gdk_pixbuf_animation_get_iter(
-	    handle->preview_animation, NULL
-    );
-
-  image = gdk_pixbuf_animation_iter_get_pixbuf(handle->preview_animation_iter);
-
-  const int delay = gdk_pixbuf_animation_iter_get_delay_time(
-    handle->preview_animation_iter
+  struct GNUNET_CHAT_File *file = (struct GNUNET_CHAT_File *) g_object_get_qdata(
+    G_OBJECT(handle->message_box),
+    handle->app->quarks.data
   );
 
-  handle->redraw_animation = g_timeout_add(
-    delay, handle_message_redraw_animation, handle
-  );
+  if (!file)
+    return FALSE;
+
+  GdkPixbuf *image = file_get_current_preview_image(file);
 
-render_image:
   if (!image)
     return FALSE;
 
@@ -258,43 +233,6 @@ render_image:
   return FALSE;
 }
 
-static void
-_clear_message_preview_data(UI_MESSAGE_Handle *handle)
-{
-  g_assert(handle);
-
-  if (handle->preview_image)
-  {
-    g_object_unref(handle->preview_image);
-    handle->preview_image = NULL;
-  }
-
-  if (handle->redraw_animation)
-  {
-    g_source_remove(handle->redraw_animation);
-    handle->redraw_animation = 0;
-  }
-
-  if (handle->preview_animation_iter)
-  {
-    g_object_unref(handle->preview_animation_iter);
-    handle->preview_animation_iter = NULL;
-  }
-
-  if (handle->preview_animation)
-  {
-    g_object_unref(handle->preview_animation);
-    handle->preview_animation = NULL;
-  }
-
-  if (handle->preview_drawing_area)
-    gtk_widget_set_size_request(
-      GTK_WIDGET(handle->preview_drawing_area),
-      -1,
-      -1
-    );
-}
-
 UI_MESSAGE_Handle*
 ui_message_new(MESSENGER_Application *app,
 	             UI_MESSAGE_Type type)
@@ -324,7 +262,7 @@ ui_message_new(MESSENGER_Application *app,
       break;
   }
 
-  handle->builder[0] = gtk_builder_new_from_resource(
+  handle->builder[0] = ui_builder_from_resource(
     application_get_resource_path(app, ui_builder_file)
   );
 
@@ -375,7 +313,11 @@ ui_message_new(MESSENGER_Application *app,
     gtk_builder_get_object(handle->builder[0], "content_box")
   );
 
-  handle->builder[1] = gtk_builder_new_from_resource(
+  handle->tag_flow_box = GTK_FLOW_BOX(
+    gtk_builder_get_object(handle->builder[0], "tag_flow_box")
+  );
+
+  handle->builder[1] = ui_builder_from_resource(
     application_get_resource_path(app, "ui/message_content.ui")
   );
 
@@ -455,6 +397,8 @@ ui_message_new(MESSENGER_Application *app,
     gtk_builder_get_object(handle->builder[1], "media_button")
   );
 
+  handle->app = app;
+
   g_signal_connect(
     handle->media_button,
     "clicked",
@@ -477,12 +421,6 @@ ui_message_new(MESSENGER_Application *app,
     gtk_builder_get_object(handle->builder[1], "message_content_box")
   ));
 
-  handle->preview_image = NULL;
-  handle->preview_animation = NULL;
-  handle->preview_animation_iter = NULL;
-
-  handle->redraw_animation = 0;
-
   return handle;
 }
 
@@ -584,6 +522,30 @@ _update_invitation_message(UI_MESSAGE_Handle *handle,
     gtk_button_clicked(handle->accept_button);
 }
 
+static void
+_update_message_with_file(UI_MESSAGE_Handle *handle,
+                          MESSENGER_Application *app,
+                          struct GNUNET_CHAT_File *file)
+{
+  g_assert(handle);
+
+  struct GNUNET_CHAT_File *prev = g_object_get_qdata(
+    G_OBJECT(handle->preview_drawing_area),
+    app->quarks.data
+  );
+
+  if (prev)
+    file_remove_widget_from_preview(file, GTK_WIDGET(handle->preview_drawing_area));
+  if (file)
+    file_add_widget_to_preview(file, GTK_WIDGET(handle->preview_drawing_area));
+
+  g_object_set_qdata(
+    G_OBJECT(handle->preview_drawing_area),
+    app->quarks.data,
+    file
+  );
+}
+
 static void
 _update_file_message(UI_MESSAGE_Handle *handle,
 		                 MESSENGER_Application *app,
@@ -607,28 +569,21 @@ _update_file_message(UI_MESSAGE_Handle *handle,
     );
 
     if ((app->settings.accept_all_files) &&
-        (!GNUNET_CHAT_file_is_downloading(file)))
+        (GNUNET_YES != GNUNET_CHAT_file_is_downloading(file)))
       autostart_download = TRUE;
 
     goto file_content;
   }
 
-  if (!(handle->preview_drawing_area))
+  if ((!(handle->preview_drawing_area)) ||
+      (GNUNET_CHAT_file_get_size(file) != GNUNET_CHAT_file_get_local_size(file)))
     goto file_progress;
 
-  const char *preview = GNUNET_CHAT_file_open_preview(file);
+  file_load_preview_image(file);
 
-  if (!preview)
-    goto file_progress;
+  GdkPixbuf *image = file_get_current_preview_image(file);
 
-  handle->preview_animation = gdk_pixbuf_animation_new_from_file(
-      preview, NULL
-  );
-
-  if (!(handle->preview_animation))
-    handle->preview_image = gdk_pixbuf_new_from_file(preview, NULL);
-
-  if ((handle->preview_animation) || (handle->preview_image))
+  if (image)
   {
     gtk_widget_set_size_request(
       GTK_WIDGET(handle->preview_drawing_area),
@@ -641,12 +596,10 @@ _update_file_message(UI_MESSAGE_Handle *handle,
       GTK_WIDGET(handle->preview_drawing_area)
     );
 
-    gtk_widget_queue_draw(GTK_WIDGET(handle->preview_drawing_area));
+    _update_message_with_file(handle, app, file);
     return;
   }
 
-  GNUNET_CHAT_file_close_preview(file);
-
   if (_message_media_supports_file_extension(filename))
   {
     gtk_image_set_from_icon_name(
@@ -713,7 +666,7 @@ ui_message_update(UI_MESSAGE_Handle *handle,
 		              MESSENGER_Application *app,
 		              const struct GNUNET_CHAT_Message *msg)
 {
-  g_assert((handle) && (app) && (msg));
+  g_assert((handle) && (app));
 
   struct GNUNET_CHAT_File *file = NULL;
   struct GNUNET_CHAT_Invitation *invitation = NULL;
@@ -780,13 +733,134 @@ ui_message_set_contact(UI_MESSAGE_Handle *handle,
 }
 
 void
-ui_message_delete(UI_MESSAGE_Handle *handle)
+ui_message_add_tag(UI_MESSAGE_Handle *handle,
+                   MESSENGER_Application *app,
+                   const struct GNUNET_CHAT_Message *tag_message)
 {
-  g_assert(handle);
+  g_assert((handle) && (app) && (tag_message));
 
+  if ((GNUNET_CHAT_KIND_TAG != GNUNET_CHAT_message_get_kind(tag_message)) ||
+      (GNUNET_CHAT_message_get_target(tag_message) != handle->msg))
+    return;
+
+  const char *tag_value = GNUNET_CHAT_message_get_text(tag_message);
+
+  if ((!tag_value) || (!*tag_value))
+    return;
+
+  UI_TAG_Handle *tag = ui_tag_new(app);
+  ui_tag_set_message(tag, app, tag_message);
+
+  gtk_container_add(GTK_CONTAINER(handle->tag_flow_box), GTK_WIDGET(tag->tag_label));
+  gtk_widget_show_all(GTK_WIDGET(tag->tag_label));
+}
+
+static void
+_remove_tag_from_message(UI_MESSAGE_Handle *handle,
+                         MESSENGER_Application *app,
+                         GtkWidget *child)
+{
+  g_assert((handle) && (app) && (child));
+
+  GList *items = gtk_container_get_children(GTK_CONTAINER(child));
+  UI_TAG_Handle *tag = NULL;
+
+  if (items)
+  {
+    GtkLabel *tag_label = GTK_LABEL(items->data);
+
+    tag = g_object_get_qdata(
+      G_OBJECT(tag_label),
+      app->quarks.ui
+    );
+
+    g_list_free(items);
+  }
+
+  gtk_container_remove(GTK_CONTAINER(handle->tag_flow_box), child);
+
+  ui_tag_delete(tag);
+  gtk_widget_destroy(child);
+}
+
+void
+ui_message_remove_tag(UI_MESSAGE_Handle *handle,
+                      MESSENGER_Application *app,
+                      const struct GNUNET_CHAT_Message *tag_message)
+{
+  g_assert((handle) && (app) && (tag_message));
+
+  if ((GNUNET_CHAT_KIND_TAG != GNUNET_CHAT_message_get_kind(tag_message)) ||
+      (GNUNET_CHAT_message_get_target(tag_message) != handle->msg))
+    return;
+  
+  GList *children = gtk_container_get_children(GTK_CONTAINER(handle->tag_flow_box));
+
+  if (!children)
+    return;
+
+  GtkWidget *removable = NULL;
+
+  GList *list = children;
+  while (list)
+  {
+    GtkFlowBoxChild *child = GTK_FLOW_BOX_CHILD(list->data);
+    GList *items = gtk_container_get_children(GTK_CONTAINER(child));
+
+    if (items)
+    {
+      GtkLabel *tag_label = GTK_LABEL(items->data);
+
+      const struct GNUNET_CHAT_Message *msg = g_object_get_qdata(
+        G_OBJECT(tag_label),
+        app->quarks.data
+      );
+
+      if (tag_message == msg)
+        removable = GTK_WIDGET(child);
+
+      g_list_free(items);
+    }
+
+    list = list->next;
+
+    if (removable)
+      break;
+  }
+
+  if (children)
+    g_list_free(children);
+
+  if (!removable)
+    return;
+
+  _remove_tag_from_message(handle, app, removable);
+}
+
+void
+ui_message_delete(UI_MESSAGE_Handle *handle,
+                  MESSENGER_Application *app)
+{
+  g_assert((handle) && (app));
+
+  _update_message_with_file(handle, app, NULL);
   ui_message_set_contact(handle, NULL);
 
-  _clear_message_preview_data(handle);
+  GList *children = gtk_container_get_children(GTK_CONTAINER(handle->tag_flow_box));
+
+  GList *list = children;
+  while (list)
+  {
+    GtkWidget *child = GTK_WIDGET(list->data);
+
+    if (child)
+      _remove_tag_from_message(handle, app, child);
+
+    list = list->next;
+  }
+
+  if (children)
+    g_list_free(children);
 
   g_object_unref(handle->builder[1]);
   g_object_unref(handle->builder[0]);
