@@ -28,6 +28,7 @@
 
 #include "account_entry.h"
 #include "chat_entry.h"
+#include "chat_title.h"
 #include "contacts.h"
 #include "message.h"
 #include "new_contact.h"
@@ -36,9 +37,27 @@
 #include "new_platform.h"
 #include "settings.h"
 
+#include "../account.h"
 #include "../application.h"
 #include "../ui.h"
 
+static void
+handle_profile_button_click(UNUSED GtkButton* button,
+				                             gpointer user_data)
+{
+  g_assert(user_data);
+
+  UI_MESSENGER_Handle *handle = (UI_MESSENGER_Handle*) user_data;
+  MESSENGER_Application *app = handle->app;
+
+  hdy_flap_set_reveal_flap(handle->flap_user_details, FALSE);
+
+  ui_contact_info_dialog_init(app, &(app->ui.contact_info));
+  ui_contact_info_dialog_update(&(app->ui.contact_info), NULL, FALSE);
+
+  gtk_widget_show(GTK_WIDGET(app->ui.contact_info.dialog));
+}
+
 static gboolean
 _flap_user_details_reveal_switch(gpointer user_data)
 {
@@ -53,7 +72,7 @@ _flap_user_details_reveal_switch(gpointer user_data)
     hdy_flap_set_reveal_flap(flap, TRUE);
   }
 
-  gtk_widget_set_sensitive(GTK_WIDGET(handle->chats_search), TRUE);
+  gtk_widget_set_sensitive(GTK_WIDGET(handle->chats_search_entry), TRUE);
   gtk_widget_set_sensitive(GTK_WIDGET(handle->chats_listbox), TRUE);
   return FALSE;
 }
@@ -66,9 +85,9 @@ handle_user_details_via_button_click(UNUSED GtkButton* button,
 
   UI_MESSENGER_Handle *handle = (UI_MESSENGER_Handle*) user_data;
 
-  gtk_widget_set_sensitive(GTK_WIDGET(handle->chats_search), FALSE);
+  gtk_widget_set_sensitive(GTK_WIDGET(handle->chats_search_entry), FALSE);
   gtk_widget_set_sensitive(GTK_WIDGET(handle->chats_listbox), FALSE);
-  g_idle_add(
+  util_idle_add(
     G_SOURCE_FUNC(_flap_user_details_reveal_switch),
     handle
   );
@@ -260,18 +279,17 @@ handle_chats_listbox_row_activated(UNUSED GtkListBox* listbox,
   if ((!entry) || (!(entry->chat)) || (!(entry->chat->chat_box)))
     return;
 
-  GtkStack *stack = app->ui.messenger.chats_stack;
-  HdyLeaflet *leaflet = app->ui.messenger.leaflet_chat;
-
-  GList *children = gtk_container_get_children(GTK_CONTAINER(leaflet));
+  UI_MESSENGER_Handle *ui = &(app->ui.messenger);
+  GList *children = gtk_container_get_children(GTK_CONTAINER(ui->leaflet_chat));
 
   if ((children) && (children->next))
-    hdy_leaflet_set_visible_child(leaflet, GTK_WIDGET(children->next->data));
+    hdy_leaflet_set_visible_child(ui->leaflet_chat, GTK_WIDGET(children->next->data));
 
   if (children)
     g_list_free(children);
 
-  gtk_stack_set_visible_child(stack, entry->chat->chat_box);
+  gtk_stack_set_visible_child(ui->chats_stack, entry->chat->chat_box);
+  gtk_stack_set_visible_child(ui->chat_title_stack, entry->chat->title->chat_title_box);
 }
 
 static gint
@@ -322,7 +340,7 @@ handle_chats_listbox_filter_func(GtkListBoxRow *row,
     return TRUE;
 
   const gchar *filter = gtk_entry_get_text(
-    GTK_ENTRY(app->ui.messenger.chats_search)
+    GTK_ENTRY(app->ui.messenger.chats_search_entry)
   );
 
   if (!filter)
@@ -343,6 +361,23 @@ handle_chats_listbox_filter_func(GtkListBoxRow *row,
   return g_str_match_string(filter, title, TRUE);
 }
 
+static void
+handle_search_button_click(UNUSED GtkButton *button,
+			                     gpointer user_data)
+{
+  g_assert(user_data);
+
+  UI_MESSENGER_Handle *handle = (UI_MESSENGER_Handle*) user_data;
+
+  if (handle->search_box == gtk_stack_get_visible_child(handle->chats_title_stack))
+  {
+    gtk_stack_set_visible_child(handle->chats_title_stack, handle->title_box);
+    gtk_entry_set_text(GTK_ENTRY(handle->chats_search_entry), "");
+  }
+  else
+    gtk_stack_set_visible_child(handle->chats_title_stack, handle->search_box);
+}
+
 static void
 handle_chats_search_changed(UNUSED GtkSearchEntry *search,
 			                      gpointer user_data)
@@ -370,7 +405,8 @@ handle_main_window_destroy(UNUSED GtkWidget *window,
 #endif
 
   ui_messenger_cleanup(&(app->ui.messenger));
-  ui_accounts_dialog_cleanup(&(app->ui.accounts));
+  
+  account_cleanup_infos();
 
   application_exit(app, MESSENGER_QUIT);
 }
@@ -384,7 +420,7 @@ ui_messenger_init(MESSENGER_Application *app,
   memset(handle, 0, sizeof(*handle));
   handle->app = app;
 
-  handle->builder = gtk_builder_new_from_resource(
+  handle->builder = ui_builder_from_resource(
     application_get_resource_path(app, "ui/messenger.ui")
   );
 
@@ -412,16 +448,32 @@ ui_messenger_init(MESSENGER_Application *app,
     1100, 700
   );
 
+  handle->leaflet_title = HDY_LEAFLET(
+    gtk_builder_get_object(handle->builder, "leaflet_title")
+  );
+
   handle->leaflet_chat = HDY_LEAFLET(
     gtk_builder_get_object(handle->builder, "leaflet_chat")
   );
 
+  g_object_bind_property(
+    handle->leaflet_chat,
+    "visible_child_name",
+    handle->leaflet_title,
+    "visible_child_name",
+    G_BINDING_SYNC_CREATE
+  );
+
   handle->flap_user_details = HDY_FLAP(
     gtk_builder_get_object(handle->builder, "flap_user_details")
   );
 
-  handle->title_bar = HDY_HEADER_BAR(
-    gtk_builder_get_object(handle->builder, "title_bar")
+  handle->nav_bar = HDY_HEADER_BAR(
+    gtk_builder_get_object(handle->builder, "nav_bar")
+  );
+
+  handle->main_bar = HDY_HEADER_BAR(
+    gtk_builder_get_object(handle->builder, "main_bar")
   );
 
   GtkLabel* application_name_label = GTK_LABEL(
@@ -432,16 +484,6 @@ ui_messenger_init(MESSENGER_Application *app,
     gtk_builder_get_object(handle->builder, "application-version-label")
   );
 
-  hdy_header_bar_set_title(
-    handle->title_bar,
-    MESSENGER_APPLICATION_TITLE
-  );
-
-  hdy_header_bar_set_subtitle(
-    handle->title_bar,
-    MESSENGER_APPLICATION_SUBTITLE
-  );
-
   gtk_label_set_text(
     application_name_label,
     MESSENGER_APPLICATION_APPNAME
@@ -455,20 +497,21 @@ ui_messenger_init(MESSENGER_Application *app,
   g_object_bind_property(
     handle->leaflet_chat,
     "folded",
-    handle->title_bar,
+    handle->main_bar,
     "show-close-button",
     G_BINDING_INVERT_BOOLEAN
   );
+  
+  handle->profile_button = GTK_BUTTON(
+    gtk_builder_get_object(handle->builder, "profile_button")
+  );
 
-  if (app->settings.mobile_design)
-    g_object_bind_property(
-      handle->leaflet_chat,
-      "folded",
-      handle->title_bar,
-      "visible",
-      G_BINDING_SYNC_CREATE |
-      G_BINDING_INVERT_BOOLEAN
-    );
+  g_signal_connect(
+    handle->profile_button,
+    "clicked",
+    G_CALLBACK(handle_profile_button_click),
+    handle
+  );
 
   handle->profile_avatar = HDY_AVATAR(
     gtk_builder_get_object(handle->builder, "profile_avatar")
@@ -604,6 +647,26 @@ ui_messenger_init(MESSENGER_Application *app,
     app
   );
 
+  handle->chats_title_stack = GTK_STACK(
+    gtk_builder_get_object(handle->builder, "chats_title_stack")
+  );
+
+  handle->title_box = GTK_WIDGET(
+    gtk_builder_get_object(handle->builder, "title_box")
+  );
+
+  handle->search_box = GTK_WIDGET(
+    gtk_builder_get_object(handle->builder, "search_box")
+  );
+
+  g_object_bind_property(
+    handle->leaflet_chat,
+    "folded",
+    handle->nav_bar,
+    "hexpand",
+    G_BINDING_SYNC_CREATE
+  );
+
   handle->user_details_button = GTK_BUTTON(
     gtk_builder_get_object(handle->builder, "user_details_button")
   );
@@ -615,8 +678,31 @@ ui_messenger_init(MESSENGER_Application *app,
     handle
   );
 
-  handle->chats_search = GTK_SEARCH_ENTRY(
-    gtk_builder_get_object(handle->builder, "chats_search")
+  handle->chats_search_button = GTK_BUTTON(
+    gtk_builder_get_object(handle->builder, "chats_search_button")
+  );
+
+  g_signal_connect(
+    handle->chats_search_button,
+    "clicked",
+    G_CALLBACK(handle_search_button_click),
+    handle
+  );
+
+  handle->chats_search_entry = GTK_SEARCH_ENTRY(
+    gtk_builder_get_object(handle->builder, "chats_search_entry")
+  );
+
+  handle->search_icon_stack = GTK_STACK(
+    gtk_builder_get_object(handle->builder, "search_icon_stack")
+  );
+
+  g_object_bind_property(
+    handle->chats_title_stack,
+    "visible_child_name",
+    handle->search_icon_stack,
+    "visible_child_name",
+    G_BINDING_SYNC_CREATE
   );
 
   handle->chats_listbox = GTK_LIST_BOX(
@@ -638,7 +724,7 @@ ui_messenger_init(MESSENGER_Application *app,
   );
 
   g_signal_connect(
-    handle->chats_search,
+    handle->chats_search_entry,
     "search-changed",
     G_CALLBACK(handle_chats_search_changed),
     handle->chats_listbox
@@ -659,6 +745,10 @@ ui_messenger_init(MESSENGER_Application *app,
     gtk_builder_get_object(handle->builder, "no_chat_box")
   );
 
+  handle->chat_title_stack = GTK_STACK(
+    gtk_builder_get_object(handle->builder, "chat_title_stack")
+  );
+
   g_signal_connect(
     handle->main_window,
     "destroy",
@@ -677,12 +767,9 @@ _messenger_iterate_accounts(void *cls,
   MESSENGER_Application *app = (MESSENGER_Application*) cls;
   UI_MESSENGER_Handle *ui = &(app->ui.messenger);
 
-  const char *name = GNUNET_CHAT_account_get_name(account);
-
   UI_ACCOUNT_ENTRY_Handle *entry = ui_account_entry_new(app);
 
-  ui_avatar_set_text(entry->entry_avatar, name);
-  ui_label_set_text(entry->entry_label, name);
+  ui_account_entry_set_account(entry, account);
 
   gtk_list_box_prepend(ui->accounts_listbox, entry->entry_box);
 
@@ -690,8 +777,7 @@ _messenger_iterate_accounts(void *cls,
 
   g_object_set_qdata(G_OBJECT(row), app->quarks.data, account);
 
-  if ((account == GNUNET_CHAT_get_connected(handle)) ||
-      ((app->chat.identity) && (0 == g_strcmp0(app->chat.identity, name))))
+  if (account == GNUNET_CHAT_get_connected(handle))
     gtk_widget_activate(row);
 
   ui_account_entry_delete(entry);
@@ -716,6 +802,17 @@ _clear_accounts_listbox(GtkWidget *widget,
   );
 }
 
+static gboolean
+_close_messenger_missing_account(gpointer data)
+{
+  g_assert(data);
+
+  UI_MESSENGER_Handle *handle = (UI_MESSENGER_Handle*) data;  
+
+  gtk_window_close(GTK_WINDOW(handle->main_window));
+  return FALSE;
+}
+
 void
 ui_messenger_refresh(MESSENGER_Application *app,
 		                 UI_MESSENGER_Handle *handle)
@@ -736,6 +833,16 @@ ui_messenger_refresh(MESSENGER_Application *app,
     _messenger_iterate_accounts,
     app
   );
+
+  if (gtk_list_box_get_selected_row(handle->accounts_listbox))
+    return;
+  
+  gtk_widget_hide(GTK_WIDGET(handle->main_window));
+
+  handle->account_refresh = util_idle_add(
+    G_SOURCE_FUNC(_close_messenger_missing_account),
+    handle
+  );
 }
 
 gboolean
@@ -773,10 +880,10 @@ ui_messenger_cleanup(UI_MESSENGER_Handle *handle)
     g_list_free_full(handle->chat_entries, (GDestroyNotify) ui_chat_entry_delete);
 
   if (handle->chat_selection)
-    g_source_remove(handle->chat_selection);
+    util_source_remove(handle->chat_selection);
 
   if (handle->account_refresh)
-    g_source_remove(handle->account_refresh);
+    util_source_remove(handle->account_refresh);
 
   memset(handle, 0, sizeof(*handle));
 }
