@@ -26,6 +26,7 @@
 
 #include "../application.h"
 #include "../request.h"
+#include "../ui.h"
 
 #include <pipewire/pipewire.h>
 
@@ -122,7 +123,7 @@ handle_camera_combo_box_change(GtkComboBox *widget,
 
 static void
 _disable_video_processing(UI_NEW_CONTACT_Handle *handle,
-			                    gboolean drop_pipeline)
+                          gboolean drop_pipeline)
 {
   g_assert(handle);
 
@@ -131,8 +132,10 @@ _disable_video_processing(UI_NEW_CONTACT_Handle *handle,
 
   if (handle->camera_count)
     gtk_stack_set_visible_child(handle->preview_stack, handle->fail_box);
-  else
+  else if (drop_pipeline)
     gtk_stack_set_visible_child(handle->preview_stack, handle->no_camera_box);
+  else
+    gtk_stack_set_visible_child(handle->preview_stack, handle->loading_box);
 
 skip_stack:
   if ((!(handle->pipeline)) || (!drop_pipeline))
@@ -388,9 +391,9 @@ _init_camera_pipeline(MESSENGER_Application *app,
   if (access)
 #endif
   {
-    app->pw.pending = pw_core_sync(app->pw.core, 0, 0);
+    application_pw_core_init(app);
+    application_pw_main_loop_run(app);
 
-    pw_main_loop_run(app->pw.main_loop);
     pw_map_for_each(&(app->pw.globals), iterate_global, handle);
 
     if (handle->camera_count)
@@ -433,7 +436,7 @@ ui_new_contact_dialog_init(MESSENGER_Application *app,
 
   _setup_gst_pipeline(handle);
 
-  handle->builder = gtk_builder_new_from_resource(
+  handle->builder = ui_builder_from_resource(
     application_get_resource_path(app, "ui/new_contact.ui")
   );
 
