@@ -26,6 +26,10 @@
 
 #include "../application.h"
 #include "../request.h"
+#include "../ui.h"
+
+#include "contact_entry.h"
+#include "files.h"
 
 #include <gnunet/gnunet_chat_lib.h>
 #include <gnunet/gnunet_common.h>
@@ -170,13 +174,35 @@ handle_leave_chats_button_click(UNUSED GtkButton* button,
   );
 }
 
+static void
+handle_show_files_button_click(UNUSED GtkButton* button,
+				                       gpointer user_data)
+{
+  g_assert(user_data);
+
+  UI_SETTINGS_Handle *handle = (UI_SETTINGS_Handle*) user_data;
+
+  handle->open_files = true;
+
+  gtk_window_close(GTK_WINDOW(handle->dialog));
+}
+
 static void
 handle_dialog_destroy(UNUSED GtkWidget *window,
                       gpointer user_data)
 {
   g_assert(user_data);
 
-  ui_settings_dialog_cleanup((UI_SETTINGS_Handle*) user_data);
+  MESSENGER_Application *app = (MESSENGER_Application*) user_data;
+  UI_SETTINGS_Handle *handle = &(app->ui.settings);
+
+  if (handle->open_files)
+  {
+    ui_files_dialog_init(app, &(app->ui.files));
+    gtk_widget_show(GTK_WIDGET(app->ui.files.dialog));
+  }
+
+  ui_settings_dialog_cleanup(handle);
 }
 
 static void
@@ -213,6 +239,9 @@ _count_blocked_contacts(void *cls,
 {
   g_assert((cls) && (contact));
 
+  if (GNUNET_YES == GNUNET_CHAT_contact_is_owned(contact))
+    return GNUNET_YES;
+
   guint *count = (guint*) cls;
 
   if (GNUNET_YES == GNUNET_CHAT_contact_is_blocked(contact))
@@ -221,13 +250,50 @@ _count_blocked_contacts(void *cls,
   return GNUNET_YES;
 }
 
+static enum GNUNET_GenericReturnValue
+_iterate_blocked_contacts(void *cls,
+                        UNUSED struct GNUNET_CHAT_Handle *handle,
+                        struct GNUNET_CHAT_Contact *contact)
+{
+  g_assert((cls) && (contact));
+
+  if ((GNUNET_YES == GNUNET_CHAT_contact_is_owned(contact)) ||
+      (GNUNET_YES != GNUNET_CHAT_contact_is_blocked(contact)))
+    return GNUNET_YES;
+
+  MESSENGER_Application *app = (MESSENGER_Application*) cls;
+
+  UI_CONTACT_ENTRY_Handle *entry = ui_contact_entry_new(app);
+  ui_contact_entry_set_contact(entry, contact);
+
+  gtk_list_box_prepend(
+    app->ui.settings.blocked_listbox,
+    entry->entry_box
+  );
+
+  GtkListBoxRow *row = GTK_LIST_BOX_ROW(
+    gtk_widget_get_parent(entry->entry_box)
+  );
+
+  g_object_set_qdata(G_OBJECT(row), app->quarks.data, contact);
+
+  g_object_set_qdata_full(
+    G_OBJECT(row),
+    app->quarks.ui,
+    entry,
+    (GDestroyNotify) ui_contact_entry_delete
+  );
+
+  return GNUNET_YES;
+}
+
 void
 ui_settings_dialog_init(MESSENGER_Application *app,
                         UI_SETTINGS_Handle *handle)
 {
   g_assert((app) && (handle));
 
-  handle->builder = gtk_builder_new_from_resource(
+  handle->builder = ui_builder_from_resource(
     application_get_resource_path(app, "ui/settings.ui")
   );
 
@@ -314,6 +380,14 @@ ui_settings_dialog_init(MESSENGER_Application *app,
     gtk_builder_get_object(handle->builder, "blocked_label")
   );
 
+  handle->blocked_scrolled_window = GTK_WIDGET(
+    gtk_builder_get_object(handle->builder, "blocked_scrolled_window")
+  );
+
+  handle->blocked_listbox = GTK_LIST_BOX(
+    gtk_builder_get_object(handle->builder, "blocked_listbox")
+  );
+
   guint blocked_count = 0;
   GNUNET_CHAT_iterate_contacts(
     app->chat.messenger.handle,
@@ -321,12 +395,20 @@ ui_settings_dialog_init(MESSENGER_Application *app,
     &blocked_count
   );
 
+  gtk_widget_set_size_request(
+    handle->blocked_scrolled_window,
+    0,
+    56 * (blocked_count > 3? 3 : blocked_count)
+  );
+
+  gtk_widget_set_visible(handle->blocked_scrolled_window, blocked_count > 0);
+
   GString *blocked_text = g_string_new(NULL);
   if (blocked_text)
   {
     g_string_printf(
       blocked_text,
-      "%u blocked contacts",
+      _("%u blocked contacts"),
       blocked_count
     );
 
@@ -338,6 +420,14 @@ ui_settings_dialog_init(MESSENGER_Application *app,
     g_string_free(blocked_text, TRUE);
   }
 
+  GNUNET_CHAT_iterate_contacts(
+    app->chat.messenger.handle,
+    _iterate_blocked_contacts,
+    app
+  );
+
+  gtk_list_box_invalidate_filter(handle->blocked_listbox);
+
   handle->read_receipts_switch = GTK_SWITCH(
     gtk_builder_get_object(handle->builder, "read_receipts_switch")
   );
@@ -442,6 +532,17 @@ ui_settings_dialog_init(MESSENGER_Application *app,
     &(app->settings.delete_files_delay)
   );
 
+  handle->show_files_button = GTK_BUTTON(
+    gtk_builder_get_object(handle->builder, "show_files_button")
+  );
+
+  g_signal_connect(
+    handle->show_files_button,
+    "clicked",
+    G_CALLBACK(handle_show_files_button_click),
+    handle
+  );
+
   handle->delete_files_button = GTK_BUTTON(
     gtk_builder_get_object(handle->builder, "delete_files_button")
   );
@@ -477,7 +578,7 @@ ui_settings_dialog_init(MESSENGER_Application *app,
     handle->dialog,
     "destroy",
     G_CALLBACK(handle_dialog_destroy),
-    handle
+    app
   );
 }
 
@@ -486,7 +587,8 @@ ui_settings_dialog_cleanup(UI_SETTINGS_Handle *handle)
 {
   g_assert(handle);
 
-  g_object_unref(handle->builder);
+  if (handle->builder)
+    g_object_unref(handle->builder);
 
   memset(handle, 0, sizeof(*handle));
 }
